from typing import List, Union
import math
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np

from .variational_baseparams import QAOAVariationalBaseParams
from ..ansatz_constructor import QAOADescriptor
from ..ansatz_constructor.baseparams import shapedArray


class QAOAVariationalExtendedParams(QAOAVariationalBaseParams):
    """
    QAOA parameters in their most general form with different angles for each
    operator.

    This means, that at the i-th timestep the evolution hamiltonian is given by

    .. math::

        H(t_i) = \sum_{\textrm{qubits } j} \beta_{ij} X_j
               + \sum_{\textrm{qubits } j} \gamma_{\textrm{single } ij} Z_j
               + \sum_{\textrm{qubit pairs} (jk)} \gamma_{\textrm{pair } i(jk)} Z_j Z_k

    and the complete circuit is then

    .. math::

        U = e^{-i H(t_p)} \cdots e^{-iH(t_1)}.

    Attributes
    ----------
    qaoa_descriptor: QAOADescriptor
                Specify the circuit parameters to construct circuit angles to be
                used for training
    betas_singles: list
        2D array with the gammas from above for each timestep and qubit.
        1st index goes over the timelayers, 2nd over the qubits.
    betas_pairs : list
    gammas_pairs: list
    gammas_singles: list
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        betas_singles: List[Union[float, int]],
        betas_pairs: List[Union[float, int]],
        gammas_singles: List[Union[float, int]],
        gammas_pairs: List[Union[float, int]],
    ):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_descriptor)

        self.betas_singles = betas_singles if self.mixer_1q_coeffs else []
        self.betas_pairs = betas_pairs if self.mixer_2q_coeffs else []
        self.gammas_singles = gammas_singles if self.cost_1q_coeffs else []
        self.gammas_pairs = gammas_pairs if self.cost_2q_coeffs else []

    def __repr__(self):
        string = "Extended Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "Parameters:\n"
        string += (
            "\tbetas_singles: " + str(self.betas_singles).replace("\n", ",") + "\n"
        )
        string += "\tbetas_pairs: " + str(self.betas_pairs).replace("\n", ",") + "\n"
        string += (
            "\tgammas_singles: " + str(self.gammas_singles).replace("\n", ",") + "\n"
        )
        string += "\tgammas_pairs: " + str(self.gammas_pairs).replace("\n", ",") + "\n"
        return string

    def __len__(self):
        return self.p * (
            len(self.mixer_1q_coeffs)
            + len(self.mixer_2q_coeffs)
            + len(self.cost_1q_coeffs)
            + len(self.cost_2q_coeffs)
        )

    @shapedArray
    def betas_singles(self):
        return (self.p, len(self.mixer_1q_coeffs))

    @shapedArray
    def betas_pairs(self):
        return (self.p, len(self.mixer_2q_coeffs))

    @shapedArray
    def gammas_singles(self):
        return (self.p, len(self.cost_1q_coeffs))

    @shapedArray
    def gammas_pairs(self):
        return (self.p, len(self.cost_2q_coeffs))

    @property
    def mixer_1q_angles(self):
        return 2 * (self.mixer_1q_coeffs * self.betas_singles)

    @property
    def mixer_2q_angles(self):
        return 2 * (self.mixer_2q_coeffs * self.betas_pairs)

    @property
    def cost_1q_angles(self):
        return 2 * (self.cost_1q_coeffs * self.gammas_singles)

    @property
    def cost_2q_angles(self):
        return 2 * (self.cost_2q_coeffs * self.gammas_pairs)

    def update_from_raw(self, new_values):
        self.betas_singles = np.array(new_values[: len(self.mixer_1q_coeffs) * self.p])
        self.betas_singles = self.betas_singles.reshape(
            (self.p, len(self.mixer_1q_coeffs))
        )

        new_values = new_values[len(self.betas_singles.flatten()) :]

        self.betas_pairs = np.array(new_values[: len(self.mixer_2q_coeffs) * self.p])
        self.betas_pairs = self.betas_pairs.reshape((self.p, len(self.mixer_2q_coeffs)))

        new_values = new_values[len(self.betas_pairs.flatten()) :]

        self.gammas_singles = np.array(new_values[: len(self.cost_1q_coeffs) * self.p])
        self.gammas_singles = self.gammas_singles.reshape(
            (self.p, len(self.cost_1q_coeffs))
        )

        new_values = new_values[len(self.gammas_singles.flatten()) :]

        self.gammas_pairs = np.array(new_values[: len(self.cost_2q_coeffs) * self.p])
        self.gammas_pairs = self.gammas_pairs.reshape(
            (self.p, len(self.cost_2q_coeffs))
        )

        new_values = new_values[len(self.gammas_pairs.flatten()) :]
        # PEP8 complains, but new_values could be np.array and not list!
        if len(new_values) != 0:
            raise RuntimeWarning(
                "Incorrect dimension specified for new_values"
                "to construct the new betas and new gammas"
            )

    def raw(self):
        raw_data = np.concatenate(
            (
                self.betas_singles.flatten(),
                self.betas_pairs.flatten(),
                self.gammas_singles.flatten(),
                self.gammas_pairs.flatten(),
            )
        )
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, time: float = None
    ):
        """

        Returns
        -------
        ExtendedParams
            The initial parameters according to a linear ramp for the Hamiltonian specified by
            register, terms, weights.

        Todo
        ----
        Refactor this s.t. it supers from __init__
        """
        # create evenly spaced timelayers at the centers of p intervals
        p = qaoa_descriptor.p
        if time is None:
            time = float(0.7 * p)

        dt = time / p

        n_gamma_singles = len(qaoa_descriptor.cost_single_qubit_coeffs)
        n_gamma_pairs = len(qaoa_descriptor.cost_pair_qubit_coeffs)
        n_beta_singles = len(qaoa_descriptor.mixer_single_qubit_coeffs)
        n_beta_pairs = len(qaoa_descriptor.mixer_pair_qubit_coeffs)

        betas = np.linspace(
            (dt / time) * (time * (1 - 0.5 / p)), (dt / time) * (time * 0.5 / p), p
        )
        gammas = betas[::-1]

        betas_singles = betas.repeat(n_beta_singles).reshape(p, n_beta_singles)
        betas_pairs = betas.repeat(n_beta_pairs).reshape(p, n_beta_pairs)
        gammas_singles = gammas.repeat(n_gamma_singles).reshape(p, n_gamma_singles)
        gammas_pairs = gammas.repeat(n_gamma_pairs).reshape(p, n_gamma_pairs)

        params = cls(
            qaoa_descriptor, betas_singles, betas_pairs, gammas_singles, gammas_pairs
        )
        return params

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, seed: int = None):
        """
        Returns
        -------
        ExtendedParams
            Randomly initialised ``ExtendedParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        p = qaoa_descriptor.p
        n_gamma_singles = len(qaoa_descriptor.cost_single_qubit_coeffs)
        n_gamma_pairs = len(qaoa_descriptor.cost_pair_qubit_coeffs)
        n_beta_singles = len(qaoa_descriptor.mixer_single_qubit_coeffs)
        n_beta_pairs = len(qaoa_descriptor.mixer_pair_qubit_coeffs)

        betas_singles = np.random.uniform(0, np.pi, (p, n_beta_singles))
        betas_pairs = np.random.uniform(0, np.pi, (p, n_beta_pairs))
        gammas_singles = np.random.uniform(0, np.pi, (p, n_gamma_singles))
        gammas_pairs = np.random.uniform(0, np.pi, (p, n_gamma_pairs))

        params = cls(
            qaoa_descriptor, betas_singles, betas_pairs, gammas_singles, gammas_pairs
        )
        return params

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor):
        """
        Initialise Extended parameters with empty arrays
        """

        p = qaoa_descriptor.p
        n_gamma_singles = len(qaoa_descriptor.cost_single_qubit_coeffs)
        n_gamma_pairs = len(qaoa_descriptor.cost_pair_qubit_coeffs)
        n_beta_singles = len(qaoa_descriptor.mixer_single_qubit_coeffs)
        n_beta_pairs = len(qaoa_descriptor.mixer_pair_qubit_coeffs)

        betas_singles = np.empty((p, n_beta_singles))
        betas_pairs = np.empty((p, n_beta_pairs))
        gammas_singles = np.empty((p, n_gamma_singles))
        gammas_pairs = np.empty((p, n_gamma_pairs))

        params = cls(
            qaoa_descriptor, betas_singles, betas_pairs, gammas_singles, gammas_pairs
        )
        return params

    def get_constraints(self):
        """Constraints on the parameters for constrained parameters.

        Returns
        -------
        List[Tuple]:
            A list of tuples (0, upper_boundary) of constraints on the
            parameters s.t. we are exploiting the periodicity of the cost
            function. Useful for constrained optimizers.

        """
        beta_constraints = [(0, math.pi)] * (
            len(self.betas_singles.flatten() + len(self.betas_pairs.flatten()))
        )

        beta_pair_constraints = [(0, math.pi / w) for w in self.mixer_2q_coeffs]
        beta_pair_constraints *= self.p

        beta_single_constraints = [(0, math.pi / w) for w in self.mixer_1q_coeffs]
        beta_single_constraints *= self.p

        gamma_pair_constraints = [(0, 2 * math.pi / w) for w in self.cost_2q_coeffs]
        gamma_pair_constraints *= self.p

        gamma_single_constraints = [(0, 2 * math.pi / w) for w in self.cost_1q_coeffs]
        gamma_single_constraints *= self.p

        all_constraints = (
            beta_single_constraints
            + beta_pair_constraints
            + gamma_single_constraints
            + gamma_pair_constraints
        )

        return all_constraints

    def plot(self, ax=None, **kwargs):
        list_names_ = ["betas singles", "betas pairs", "gammas singles", "gammas pairs"]
        list_values_ = [
            self.betas_singles % (2 * (np.pi)),
            self.betas_pairs % (2 * (np.pi)),
            self.gammas_singles % (2 * (np.pi)),
            self.gammas_pairs % (2 * (np.pi)),
        ]

        list_names, list_values = list_names_.copy(), list_values_.copy()

        n_pop = 0
        for i in range(len(list_values_)):
            if list_values_[i].size == 0:
                list_values.pop(i - n_pop)
                list_names.pop(i - n_pop)
                n_pop += 1

        n = len(list_values)
        p = self.p

        if ax is None:
            fig, ax = plt.subplots((n + 1) // 2, 2, figsize=(9, 9 if n > 2 else 5))

        fig.tight_layout(pad=4.0)

        for k, (name, values) in enumerate(zip(list_names, list_values)):
            i, j = k // 2, k % 2
            axes = ax[i, j] if n > 2 else ax[k]

            if values.size == p:
                axes.plot(values.T[0], marker="^", color="green", ls="", **kwargs)
                axes.set_xlabel("p", fontsize=12)
                axes.set_title(name)
                axes.xaxis.set_major_locator(MaxNLocator(integer=True))

            else:
                n_terms = values.shape[1]
                plt1 = axes.pcolor(
                    np.arange(p),
                    np.arange(n_terms),
                    values.T,
                    vmin=0,
                    vmax=2 * np.pi,
                    cmap="seismic",
                )
                axes.set_aspect(p / n_terms)
                axes.xaxis.set_major_locator(MaxNLocator(integer=True))
                axes.yaxis.set_major_locator(MaxNLocator(integer=True))
                axes.set_ylabel("terms")
                axes.set_xlabel("p")
                axes.set_title(name)

                plt.colorbar(plt1, **kwargs)

        if k == 0:
            ax[1].axis("off")
        elif k == 2:
            ax[1, 1].axis("off")



import numpy as np
from docplex.mp.model import Model
import networkx as nx
import matplotlib.pyplot as plt

from .problem import Problem
from .converters import FromDocplex2IsingModel
from .qubo import QUBO


class PortfolioOptimization(Problem):
    """
    Creates an instance of the portfolio optimization problem.
    https://en.wikipedia.org/wiki/Portfolio_optimization

    Parameters
    ----------
    mu: List[int]
        The expected return
    sigma: List[List[int]]
        The covariance matrix.
    risk_factor: float
        desired risk factor exposures
    budget: int
        maximum number of assets to buy
    penalty: float
        Penalty for the maximum budget.

    Returns
    -------
        An instance of the portfolio optimization problem.
    """

    __name__ = "portfolio_optimization"

    def __init__(self, mu, sigma, risk_factor, budget, penalty):
        # Check whether the input is valid. Number of values should match the number of weights.
        for s_i in sigma:
            if len(mu) != len(s_i):
                raise ValueError("Number of assets does not match sigma")
        self.mu = mu
        self.sigma = sigma
        self.risk_factor = risk_factor
        self.penalty = penalty
        self.num_assets = len(mu)
        self.budget = budget

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the portfolio optimization problem.

        Parameters
        ----------
        num_assets: int
            The number of assets.
        mu_bounds: tuple
            bounds of the expected return
        sigma_bounds: tuple
            bound of the covariance matrix
        risk_factor: float
            desired risk factor exposures
        seed: int
            Seed for the random number selection.

        Returns
        -------
            A random instance of the portfolio optimization problem.
        """
        num_assets = kwargs.get("num_assets", 8)
        seed = kwargs.get("seed", 1234)
        mu_bounds = kwargs.get("mu_bounds", (-1, 1))
        sigma_bounds = kwargs.get("sigma_bounds", (-0.1, 0.1))
        budget = kwargs.get("budget", 5)
        risk_factor = kwargs.get("risk_factor", 0.1)
        penalty = kwargs.get("penalty", 2)

        rng = np.random.default_rng(seed)

        mu = [
            (mu_bounds[1] - mu_bounds[0]) * rng.random() + mu_bounds[0]
            for _ in range(num_assets)
        ]
        sigma = [[0 for i in range(num_assets)] for j in range(num_assets)]
        for i in range(num_assets):
            for j in range(num_assets):
                sigma[i][j] = (
                    sigma_bounds[1] - sigma_bounds[0]
                ) * rng.random() + sigma_bounds[0]

        return PortfolioOptimization(mu, sigma, risk_factor, budget, penalty)

    @property
    def docplex_model(self):
        num_assets = self.num_assets

        # Start the docplex model with Model("name of the model")
        mdl = Model("Portfolio_Optimization")

        # Consider the number of variables as num_assets,
        # and binary set of variables that represent the stocks
        # x vector in numpy array for matrix multiplication
        x = np.array(mdl.binary_var_list(num_assets, name="asset"))

        # Specific the objective of the
        # portfolio optimization function
        objective_function = -np.array(self.mu) @ x + x.T @ np.array(self.sigma) @ x

        # For this problem it aims to maximize the profit
        # of those assets minimizing the risk of the investment
        mdl.minimize(objective_function)

        # Budget constraint
        mdl.add_constraint(mdl.sum(x) == self.budget, ctname="budget")

        return mdl

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.
        Returns
        -------
            The QUBO encoding of this problem.
        """
        cplex_model = self.docplex_model
        n_vars = cplex_model.number_of_binary_variables
        qubo_docplex = FromDocplex2IsingModel(cplex_model, multipliers=self.penalty)
        ising_model = qubo_docplex.ising_model
        return QUBO(
            n_vars,
            ising_model.terms + [[]],
            ising_model.weights + [ising_model.constant],
            self.problem_instance,
        )

    def classical_solution(self, string: bool = False):
        """
        Return the classical solution of the portfolio optimization problem
        Parameters
        ----------
        string : bool, optional
            If the solution is returned as a string. The default is False.
        Raises
        ------
        ValueError
            A flag if docplex does not find a valid solution.
        Returns
        -------
        solution : Union[str, dict]
            The classical solution of the specific problem as a string or a dict.
        """
        cplex_model = self.docplex_model
        docplex_sol = cplex_model.solve()

        if docplex_sol is None:
            raise ValueError(f"solution not found: {cplex_model.solve_details.status}")

        if string:
            solution = ""
        else:
            solution = {}
        for var in cplex_model.iter_binary_vars():
            if string:
                # round is used because sometimes docplex gives the solution close to 1 but not 1,
                # if we want to store this as an integer and use int(value) and the value is 0.999999
                # it will store a zero.
                solution += str(round(docplex_sol.get_value(var), 1))
            else:
                solution[var.name] = round(docplex_sol.get_value(var), 1)
        return solution

    def plot_solution(self, solution, ax=None):
        G = nx.Graph()
        G.add_nodes_from(range(self.num_assets))
        edge_list = []
        width = []
        for i in range(self.num_assets - 1):
            for j in range(i + 1, self.num_assets):
                G.add_edge(i, j, weight=self.sigma[i][j])
                edge_list.append((i, j))
                width.append(self.sigma[i][j])
        pos = nx.circular_layout(G)
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        graph_colors_sol = [
            "#F29D55" if int(i) else "#28C9C9" for i in solution.values()
        ]

        nx.draw(
            G,
            pos=pos,
            labels={i: k for i, k in enumerate(solution.keys())},
            node_size=900,
            node_color=graph_colors_sol,
            ax=ax,
            edgecolors="grey",
        )
        nx.draw_networkx_edges(G, pos, edgelist=edge_list, width=width, ax=ax)
        ax.plot(
            [],
            [],
            marker="o",
            linewidth=0,
            markeredgecolor="grey",
            color="#F29D55",
            label="selected",
            markersize=10,
        )
        ax.plot(
            [],
            [],
            marker="o",
            linewidth=0,
            markeredgecolor="grey",
            color="#28C9C9",
            label="non-selected",
            markersize=10,
        )

        ax.legend(loc="upper center", bbox_to_anchor=(0.5, 1.08), ncol=2)
        return fig



import unittest
from unittest.mock import Mock

import openqaoa
from openqaoa.backends.qaoa_vectorized import QAOAvectorizedBackendSimulator
import openqaoa.qaoa_components.ansatz_constructor.gates as oq_gate_mod
from openqaoa.backends.gates_vectorized import VectorizedGateApplicator
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle


class TestVectorizedGateApplicator(unittest.TestCase):
    def setUp(self):
        available_vectorized_gates_name = [
            each_name
            for each_name in dir(QAOAvectorizedBackendSimulator)
            if "apply" in each_name
        ]
        available_gates = [
            getattr(QAOAvectorizedBackendSimulator, each_name)
            for each_name in available_vectorized_gates_name
        ]

        self.available_vectorized_gates = dict(
            zip(available_vectorized_gates_name, available_gates)
        )

        self.oq_available_gates = (
            tuple(
                [each_gate for each_gate in oq_gate_mod.OneQubitGate.__subclasses__()]
            )
            + tuple(
                [
                    each_gate
                    for each_gate in oq_gate_mod.OneQubitRotationGate.__subclasses__()
                ]
            )
            + tuple(
                [each_gate for each_gate in oq_gate_mod.TwoQubitGate.__subclasses__()]
            )
            + tuple(
                [
                    each_gate
                    for each_gate in oq_gate_mod.TwoQubitRotationGate.__subclasses__()
                ]
            )
        )

        # OneQubitGate, OneQubitRotationGate, TwoQubitGate, TwoQubitRotationGate are not acceptable inputs into the function even though they are of the Gate class
        # CZ, CX, RXY and CPhase are not supported by Vectorized library
        self.vectorized_excluded_gates = [
            oq_gate_mod.OneQubitGate,
            oq_gate_mod.OneQubitRotationGate,
            oq_gate_mod.TwoQubitGate,
            oq_gate_mod.TwoQubitRotationGate,
            oq_gate_mod.X,
            oq_gate_mod.CZ,
            oq_gate_mod.CX,
            oq_gate_mod.RXY,
            oq_gate_mod.CPHASE,
        ]

    def test_gate_applicator_mapper(self):
        """
        The mapper to the gate applicator should only contain gates that
        are trivially support by the library.
        """

        for each_gate in VectorizedGateApplicator.VECTORIZED_OQ_GATE_MAPPER(
            QAOAvectorizedBackendSimulator
        ).values():
            self.assertTrue(
                each_gate.__name__.lower() in self.available_vectorized_gates.keys(),
                "{}, {}".format(
                    each_gate.__name__, self.available_vectorized_gates.keys()
                ),
            )

    def test_gate_selector(self):
        """
        This method should apply the Vectorized Gate Function based on the input OQ
        Gate object.
        """

        gate_applicator = VectorizedGateApplicator()

        oq_gate_list = tuple(
            [each_gate for each_gate in oq_gate_mod.Gate.__subclasses__()]
        )

        # Vectorized Gate Function for 1 Qubit Gates without Rotation Angle
        oq_gate_list_1q = oq_gate_list + tuple(
            [each_gate for each_gate in oq_gate_mod.OneQubitGate.__subclasses__()]
        )

        for each_gate in oq_gate_list_1q:
            if each_gate not in self.vectorized_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(applicator=None, qubit_1=None, vectorized_backend=None),
                    QAOAvectorizedBackendSimulator,
                )
                self.assertEqual(
                    self.available_vectorized_gates[
                        returned_gate._extract_mock_name.lower()
                    ],
                    returned_gate,
                )

        # Vectorized Gate Function for 1 Qubit Gates with Rotation Angle
        oq_gate_list_1qr = oq_gate_list + tuple(
            [
                each_gate
                for each_gate in oq_gate_mod.OneQubitRotationGate.__subclasses__()
            ]
        )

        for each_gate in oq_gate_list_1qr:
            if each_gate not in self.vectorized_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(applicator=None, qubit_1=None, rotation_object=None),
                    QAOAvectorizedBackendSimulator,
                )
                self.assertEqual(
                    self.available_vectorized_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

        # Vectorized Gate Function for 2 Qubit Gate without Rotation Angle
        oq_gate_list_2q = oq_gate_list + tuple(
            [each_gate for each_gate in oq_gate_mod.TwoQubitGate.__subclasses__()]
        )

        for each_gate in oq_gate_list_2q:
            if each_gate not in self.vectorized_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(applicator=None, qubit_1=None, qubit_2=None),
                    QAOAvectorizedBackendSimulator,
                )
                self.assertEqual(
                    self.available_vectorized_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

        # Vectorized Gate Function for 2 Qubit Gate with Rotation Angle
        oq_gate_list_2qr = oq_gate_list + tuple(
            [
                each_gate
                for each_gate in oq_gate_mod.TwoQubitRotationGate.__subclasses__()
            ]
        )

        for each_gate in oq_gate_list_2qr:
            if each_gate not in self.vectorized_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(
                        applicator=None,
                        qubit_1=None,
                        qubit_2=None,
                        rotation_object=None,
                    ),
                    QAOAvectorizedBackendSimulator,
                )
                self.assertEqual(
                    self.available_vectorized_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

    def test_not_supported_gates(self):
        """
        If an unsupported Gate object is passed into the apply_gate method,
        a KeyError should be raised.
        The unsupported Gate object does not exist on the mapper.
        """

        unsupported_list = [
            oq_gate_mod.CZ,
            oq_gate_mod.CX,
            oq_gate_mod.RXY,
            oq_gate_mod.CPHASE,
        ]
        qubit_1 = 0
        qubit_2 = 1
        rotation_object = None

        gate_applicator = VectorizedGateApplicator()

        for each_gate in unsupported_list:
            with self.assertRaises(KeyError):
                gate_applicator.apply_gate(
                    each_gate,
                    qubit_1,
                    qubit_2,
                    rotation_object,
                    QAOAvectorizedBackendSimulator,
                )

    def test_wrong_argument(self):
        """
        If a supported Gate object is passed into the apply_gate method with the
        incorrect set of arguments, a TypeError should be raised.
        """

        input_gates = {
            tuple(oq_gate_mod.OneQubitRotationGate.__subclasses__()): {
                "wrong_args": [0, 1, None],
                "n_qubits": 1,
            },
            tuple(oq_gate_mod.TwoQubitGate.__subclasses__()): {
                "wrong_args": [0, 1, None],
                "n_qubits": 2,
            },
            tuple(oq_gate_mod.TwoQubitRotationGate.__subclasses__()): {
                "wrong_args": [0, 1, None],
                "n_qubits": 2,
            },
        }
        gate_applicator = VectorizedGateApplicator()

        for each_gate_set, input_arguments in input_gates.items():
            for each_gate in each_gate_set:
                each_gate.n_qubits = input_arguments["n_qubits"]
                if each_gate not in self.vectorized_excluded_gates:
                    with self.assertRaises(TypeError):
                        gate_applicator.apply_gate(
                            each_gate,
                            *input_arguments["wrong_args"],
                            QAOAvectorizedBackendSimulator
                        )

    def test_wrong_n_qubits(self):
        """
        If a supported Gate object is passed into the apply_gate method with the
        n_qubits attribute that is not 1 or 2, a ValueError should be raised.
        """

        input_gates = {
            tuple(oq_gate_mod.OneQubitRotationGate.__subclasses__()): {
                "args": [0, None],
                "n_qubits": 3,
            }
        }
        gate_applicator = VectorizedGateApplicator()

        for each_gate_set, input_arguments in input_gates.items():
            for each_gate in each_gate_set:
                each_gate.n_qubits = input_arguments["n_qubits"]
                if each_gate not in self.vectorized_excluded_gates:
                    with self.assertRaises(ValueError):
                        gate_applicator.apply_gate(
                            each_gate,
                            *input_arguments["args"],
                            QAOAvectorizedBackendSimulator
                        )
                    break

    def test_static_methods_1q(self):
        """
        Checks that the static method, apply_1q_rotation_gate, apply the correct
        gate function.
        This method is check directly as there are currently no OQ Gate objects
        that call this method through apply_gate.
        """

        gate_applicator = VectorizedGateApplicator()
        QAOAvectorizedBackendSimulator.apply_x = Mock()
        gate_applicator.apply_1q_fixed_gate(QAOAvectorizedBackendSimulator.apply_x, 0)

        QAOAvectorizedBackendSimulator.apply_x.assert_called_with(0)

    def test_static_methods_1qr(self):
        """
        Checks that the static method, apply_1q_rotation_gate, apply the correct
        gate function.
        """

        gate_applicator = VectorizedGateApplicator()
        input_angle = 1
        rot_obj = RotationAngle(lambda x: x, None, input_angle)

        each_sub_gate = [
            each_gate for each_gate in oq_gate_mod.OneQubitRotationGate.__subclasses__()
        ]

        for each_gate in each_sub_gate:
            mock_vectorized_backend = Mock()
            gate_applicator.apply_gate(
                each_gate(gate_applicator, 0, rot_obj),
                0,
                rot_obj,
                mock_vectorized_backend,
            )

            method_name = VectorizedGateApplicator.VECTORIZED_OQ_GATE_MAPPER(
                mock_vectorized_backend
            )[each_gate.__name__]._extract_mock_name()[5:]
            # Check that the method is called with the right arguments
            getattr(mock_vectorized_backend, method_name).assert_called_with(
                0, rot_obj.rotation_angle
            )

    def test_static_methods_2q(self):
        """
        Checks that the static method, apply_2q_fixed_gate, apply the correct
        gate function.
        """

        gate_applicator = VectorizedGateApplicator()

        each_sub_gate = [
            each_gate for each_gate in oq_gate_mod.TwoQubitGate.__subclasses__()
        ]

        for each_gate in each_sub_gate:
            if each_gate not in self.vectorized_excluded_gates:
                mock_vectorized_backend = Mock()
                gate_applicator.apply_gate(
                    each_gate(gate_applicator, 0, 1), 0, 1, mock_vectorized_backend
                )

                method_name = VectorizedGateApplicator.VECTORIZED_OQ_GATE_MAPPER(
                    mock_vectorized_backend
                )[each_gate.__name__]._extract_mock_name()[5:]
                # Check that the method is called with the right arguments
                getattr(mock_vectorized_backend, method_name).assert_called_with(0, 1)

    def test_static_methods_2qr(self):
        """
        Checks that the static method, apply_2q_rotation_gate, apply the correct
        gate function.
        """

        gate_applicator = VectorizedGateApplicator()
        input_angle = 1
        rot_obj = RotationAngle(lambda x: x, None, input_angle)

        each_sub_gate = [
            each_gate for each_gate in oq_gate_mod.TwoQubitRotationGate.__subclasses__()
        ]

        for each_gate in each_sub_gate:
            if each_gate not in self.vectorized_excluded_gates:
                mock_vectorized_backend = Mock()
                gate_applicator.apply_gate(
                    each_gate(gate_applicator, 0, 1, rot_obj),
                    0,
                    1,
                    rot_obj,
                    mock_vectorized_backend,
                )

                method_name = VectorizedGateApplicator.VECTORIZED_OQ_GATE_MAPPER(
                    mock_vectorized_backend
                )[each_gate.__name__]._extract_mock_name()[5:]
                # Check that the method is called with the right arguments
                getattr(mock_vectorized_backend, method_name).assert_called_with(
                    0, 1, rot_obj.rotation_angle
                )


if __name__ == "__main__":
    unittest.main()



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import ShortestPath, QUBO


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestShortestPath(unittest.TestCase):
    """Tests for ShortestPath class"""

    def test_shortestpath_terms_weights_constant(self):
        """Test terms,weights,constant of QUBO generated by Shortest Path class"""

        sp_terms = [
            [0],
            [1],
            [2],
            [3],
            [1],
            [1, 2],
            [2, 1],
            [2],
            [2],
            [2, 3],
            [3, 2],
            [3],
            [0],
            [0, 1],
            [1],
            [1, 3],
            [0, 3],
            [3, 1],
            [3],
        ]
        sp_weights = [1, 1, 1, 1, -1, 1, 1, -1, -1, 1, 1, -1, 4, -4, 1, 1, -4, 1, 1]
        conv_sp_terms = [
            [1, 2],
            [2, 1],
            [2, 3],
            [3, 2],
            [0, 1],
            [1, 3],
            [0, 3],
            [3, 1],
            [0],
            [1],
            [2],
            [3],
            [],
        ]
        conv_sp_weights = [
            0.25,
            0.25,
            0.25,
            0.25,
            -1.0,
            0.25,
            -1.0,
            0.25,
            -0.5,
            -0.5,
            -0.5,
            -0.5,
            2.5,
        ]
        sp_qubo_terms = [[1, 2], [2, 3], [0, 1], [1, 3], [0, 3], [0], [1], [2], [3]]
        sp_qubo_weights = [0.5, 0.5, -1.0, 0.5, -1.0, -0.5, -0.5, -0.5, -0.5]
        sp_qubo_constant = 2.5

        gr = nx.generators.fast_gnp_random_graph(3, 1, seed=1234)
        for u, v in gr.edges():
            gr.edges[u, v]["weight"] = 1
        for w in gr.nodes():
            gr.nodes[w]["weight"] = 1
        source, dest = 0, 2
        sp = ShortestPath(gr, source, dest)
        n_variables = sp.G.number_of_nodes() + sp.G.number_of_edges() - 2
        bin_terms, bin_weights = sp.terms_and_weights()
        terms, weights = QUBO.convert_qubo_to_ising(n_variables, bin_terms, bin_weights)
        qubo = sp.qubo

        self.assertTrue(terms_list_equality(bin_terms, sp_terms))
        self.assertEqual(list(bin_weights), sp_weights)
        self.assertTrue(terms_list_equality(terms, conv_sp_terms))
        self.assertEqual(list(weights), conv_sp_weights)
        self.assertTrue(terms_list_equality(sp_qubo_terms, qubo.terms))
        self.assertEqual(sp_qubo_weights, qubo.weights)
        self.assertEqual(sp_qubo_constant, qubo.constant)

    def test_shortestpath_random_instance(self):
        """Test random instance method of Shortest Path problem class"""
        sp_rand_terms = [[1, 2], [2, 3], [0, 1], [1, 3], [0, 3], [0], [1], [2], [3]]
        sp_rand_weights = [0.5, 0.5, -1.0, 0.5, -1.0, -0.5, -0.5, -0.5, -0.5]
        sp_rand_constant = 2.5

        gr = nx.generators.fast_gnp_random_graph(3, 1, seed=1234)
        for u, v in gr.edges():
            gr.edges[u, v]["weight"] = 1.0
        for w in gr.nodes():
            gr.nodes[w]["weight"] = 1.0
        sp_prob = ShortestPath.random_instance(
            n_nodes=3, edge_probability=1, seed=1234, source=0, dest=2
        ).qubo

        self.assertTrue(terms_list_equality(sp_rand_terms, sp_prob.terms))
        self.assertEqual(sp_rand_weights, sp_prob.weights)
        self.assertEqual(sp_rand_constant, sp_prob.constant)

        self.assertEqual(sp_prob.terms, ShortestPath(gr, 0, 2).qubo.terms)
        self.assertEqual(sp_prob.weights, ShortestPath(gr, 0, 2).qubo.weights)
        self.assertEqual(sp_prob.constant, ShortestPath(gr, 0, 2).qubo.constant)

    def test_assertion_error(self):
        def test_assertion_fn():
            n_row = 1
            n_col = 1

            G = nx.triangular_lattice_graph(n_row, n_col)
            G = nx.convert_node_labels_to_integers(G)
            G.remove_edges_from(nx.selfloop_edges(G))

            node_weights = np.round(np.random.rand(len(G.nodes())), 3)
            edge_weights = np.round(np.random.rand(len(G.edges())), 3)

            node_dict = dict(zip(list(G.nodes()), node_weights))
            edge_dict = dict(zip(list(G.edges()), edge_weights))

            nx.set_edge_attributes(G, values=edge_dict, name="weight")
            nx.set_node_attributes(G, values=node_dict, name="weight")

            shortest_path_problem = ShortestPath(G, 0, -1)
            shortest_path_qubo = shortest_path_problem.qubo

        self.assertRaises(Exception, test_assertion_fn)


if __name__ == "__main__":
    unittest.main()



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import KColor


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestKColor(unittest.TestCase):
    """Tests for the KColor class"""

    def test_kcolor_terms_weights_constant(self):
        """Test that KColor creates a correct Docplex model from the provided graph"""

        gr = nx.generators.random_graphs.fast_gnp_random_graph(n=3, p=0.8, seed=1234)
        k = 3

        kcolor_prob = KColor(gr, k)
        mdl = kcolor_prob.docplex_model
        qubo = kcolor_prob.qubo
        terms = [
            [3, 6],
            [4, 7],
            [8, 5],
            [0, 1],
            [0, 2],
            [1, 2],
            [3, 4],
            [3, 5],
            [4, 5],
            [6, 7],
            [8, 6],
            [8, 7],
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
        ]
        weights = [
            1.0,
            1.0,
            1.0,
            6.5,
            6.5,
            6.5,
            6.5,
            6.5,
            6.5,
            6.5,
            6.5,
            6.5,
            -6.5,
            -6.5,
            -6.5,
            -7.5,
            -7.5,
            -7.5,
            -7.5,
            -7.5,
            -7.5,
        ]
        constant = 42.0
        # Verify the number of variables
        num_vertices = gr.number_of_nodes()
        num_colors = k
        expected_num_vars = num_vertices * num_colors
        self.assertEqual(mdl.number_of_variables, expected_num_vars)

        # Verify the constraints
        expected_num_constraints = num_vertices
        self.assertEqual(mdl.number_of_constraints, expected_num_constraints)

        # Verify the qubo
        self.assertEqual(qubo.terms, terms)
        self.assertEqual(qubo.weights, weights)
        self.assertEqual(qubo.constant, constant)

    def test_kcolor_random_problem(self):
        """Test KColor random instance method"""
        seed = 1234
        n_nodes = 10
        edge_probability = 0.8
        k = 4
        np.random.seed(seed)
        kcolor_manual_prob = KColor.random_instance(
            n_nodes=n_nodes, edge_probability=edge_probability, k=k, seed=seed
        )
        kcolor_random_prob = KColor(
            nx.generators.random_graphs.fast_gnp_random_graph(
                n=n_nodes, p=edge_probability, seed=seed
            ),
            k,
        )

        # Verify the graph is the same
        self.assertEqual(kcolor_manual_prob.G.nodes(), kcolor_random_prob.G.nodes())
        self.assertEqual(kcolor_manual_prob.G.edges(), kcolor_random_prob.G.edges())

        # Verify the number of colors
        self.assertEqual(k, kcolor_manual_prob.k)

        # Verify the Docplex models are the same
        self.assertEqual(
            kcolor_manual_prob.docplex_model.export_to_string(),
            kcolor_random_prob.docplex_model.export_to_string(),
        )

    def test_kcolor_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {"node1": 1, "node2": 2}, np.array([1, 2])]

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                KColor(G=each_graph, k=3)
            self.assertEqual(
                "Input problem graph must be a networkx Graph.", str(e.exception)
            )

    def test_kcolor_classical_sol(self):
        """Test the k-color random instance method classical solution"""

        seed = 1234
        n_nodes = 10
        edge_probability = 0.1
        k = 3

        np.random.seed(seed)
        kcolor_sol = KColor.random_instance(
            n_nodes=n_nodes, edge_probability=edge_probability, k=k, seed=seed
        ).classical_solution(string=True)
        # Verify the solution size
        self.assertTrue(len(kcolor_sol) == n_nodes * k)

        # Verify each vertex has just a color
        for i in range(n_nodes):
            self.assertTrue(kcolor_sol[i * k : (i + 1) * k].count("1") == 1)

    def test_kcolor_plot(self):
        """Test k-color random instance method"""
        import matplotlib.pyplot as plt

        seed = 1234
        n_nodes = 10
        edge_probability = 0.7
        k = 3

        kcolor_random_prob = KColor.random_instance(
            n_nodes=n_nodes, edge_probability=edge_probability, k=k, seed=seed
        )
        sol = kcolor_random_prob.classical_solution()
        fig, ax = plt.subplots()
        kcolor_random_prob.plot_solution(sol, ax=ax)

        self.assertTrue(isinstance(ax, plt.Axes))


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""This module contains methods to expand the matrix representation of an operator
to a higher hilbert space with re-ordered wires."""
import copy
from functools import reduce
from typing import Generator, Tuple

import numpy as np
from scipy.sparse import csr_matrix, eye, issparse, kron

from openqaoa.optimizers import (
    pennylane as qml,
)  # changed from the original PennyLane code

Wires = None


def expand_matrix(base_matrix, wires, wire_order=None, sparse_format="csr"):
    """Re-express a base matrix acting on a subspace defined by a set of wire labels
    according to a global wire order.

    Args:
        base_matrix (tensor_like): base matrix to expand
        wires (Iterable): wires determining the subspace that base matrix acts on; a base matrix of
            dimension :math:`2^n` acts on a subspace of :math:`n` wires
        wire_order (Iterable): global wire order, which has to contain all wire labels in ``wires``, but can also
            contain additional labels
        sparse_format (str): if the base matrix is a scipy sparse matrix then this is the string representing the
            preferred scipy sparse matrix format to cast the expanded matrix too

    Returns:
        tensor_like: expanded matrix

    **Example**

    If the wire order is ``None`` or identical to ``wires``, the original matrix gets returned:

    >>> base_matrix = np.array([[1, 2, 3, 4],
    ...                         [5, 6, 7, 8],
    ...                         [9, 10, 11, 12],
    ...                         [13, 14, 15, 16]])
    >>> print(expand_matrix(base_matrix, wires=[0, 2], wire_order=[0, 2]))
    [[ 1  2  3  4]
     [ 5  6  7  8]
     [ 9 10 11 12]
     [13 14 15 16]]
    >>> print(expand_matrix(base_matrix, wires=[0, 2]))
    [[ 1  2  3  4]
     [ 5  6  7  8]
     [ 9 10 11 12]
     [13 14 15 16]]

    If the wire order is a permutation of ``wires``, the entries of the base matrix get permuted:

    >>> print(expand_matrix(base_matrix, wires=[0, 2], wire_order=[2, 0]))
    [[ 1  3  2  4]
     [ 9 11 10 12]
     [ 5  7  6  8]
     [13 15 14 16]]

    If the wire order contains wire labels not found in ``wires``, the matrix gets expanded:

    >>> print(expand_matrix(base_matrix, wires=[0, 2], wire_order=[0, 1, 2]))
    [[ 1  2  0  0  3  4  0  0]
     [ 5  6  0  0  7  8  0  0]
     [ 0  0  1  2  0  0  3  4]
     [ 0  0  5  6  0  0  7  8]
     [ 9 10  0  0 11 12  0  0]
     [13 14  0  0 15 16  0  0]
     [ 0  0  9 10  0  0 11 12]
     [ 0  0 13 14  0  0 15 16]]

    The method works with tensors from all autodifferentiation frameworks, for example:

    >>> base_matrix_torch = torch.tensor([[1., 2.],
    ...                                   [3., 4.]], requires_grad=True)
    >>> res = expand_matrix(base_matrix_torch, wires=["b"], wire_order=["a", "b"])
    >>> type(res)
    torch.Tensor
    >>> res.requires_grad
    True

    The method words with scipy sparse matrices, for example:

    >>> from scipy import sparse
    >>> mat = sparse.csr_matrix([[0, 1], [1, 0]])
    >>> qml.math.expand_matrix(mat, wires=[1], wire_order=[0,1]).toarray()
    array([[0., 1., 0., 0.],
           [1., 0., 0., 0.],
           [0., 0., 0., 1.],
           [0., 0., 1., 0.]])

    """

    if (wire_order is None) or (wire_order == wires):
        return base_matrix

    interface = qml.math.get_interface(base_matrix)  # pylint: disable=protected-access
    if interface == "scipy" and issparse(base_matrix):
        return _sparse_expand_matrix(
            base_matrix, wires, wire_order, format=sparse_format
        )

    wire_order = qml.wires.Wires(wire_order)
    n = len(wires)
    shape = qml.math.shape(base_matrix)
    batch_dim = shape[0] if len(shape) == 3 else None

    # operator's wire positions relative to wire ordering
    op_wire_pos = wire_order.indices(wires)

    identity = qml.math.reshape(
        qml.math.eye(2 ** len(wire_order), like=interface), [2] * (len(wire_order) * 2)
    )
    # The first axis entries are range(n, 2n) for batch_dim=None and range(n+1, 2n+1) else
    axes = (list(range(-n, 0)), op_wire_pos)

    # reshape op.matrix()
    op_matrix_interface = qml.math.convert_like(base_matrix, identity)
    shape = [batch_dim] + [2] * (n * 2) if batch_dim else [2] * (n * 2)
    mat_op_reshaped = qml.math.reshape(op_matrix_interface, shape)
    mat_tensordot = qml.math.tensordot(
        mat_op_reshaped, qml.math.cast_like(identity, mat_op_reshaped), axes
    )

    unused_idxs = [idx for idx in range(len(wire_order)) if idx not in op_wire_pos]
    # permute matrix axes to match wire ordering
    perm = op_wire_pos + unused_idxs
    sources = wire_order.indices(wire_order)
    if batch_dim:
        perm = [p + 1 for p in perm]
        sources = [s + 1 for s in sources]

    mat = qml.math.moveaxis(mat_tensordot, sources, perm)
    shape = (
        [batch_dim] + [2 ** len(wire_order)] * 2
        if batch_dim
        else [2 ** len(wire_order)] * 2
    )
    mat = qml.math.reshape(mat, shape)

    return mat


def reduce_matrices(
    mats_and_wires_gen: Generator[Tuple[np.ndarray, Wires], None, None],
    reduce_func: callable,
) -> Tuple[np.ndarray, Wires]:
    """Apply the given ``reduce_func`` cumulatively to the items of the ``mats_and_wires_gen``
    generator, from left to right, so as to reduce the sequence to a tuple containing a single
    matrix and the wires it acts on.

    Args:
        mats_and_wires_gen (Generator): generator of tuples containing the matrix and the wires of
            each operator
        reduce_func (callable): function used to reduce the sequence of operators

    Returns:
        Tuple[tensor, Wires]: a tuple containing the reduced matrix and the wires it acts on
    """

    def expand_and_reduce(
        op1_tuple: Tuple[np.ndarray, Wires], op2_tuple: Tuple[np.ndarray, Wires]
    ):
        mat1, wires1 = op1_tuple
        mat2, wires2 = op2_tuple
        expanded_wires = wires1 + wires2
        mat1 = qml.math.expand_matrix(mat1, wires1, wire_order=expanded_wires)
        mat2 = qml.math.expand_matrix(mat2, wires2, wire_order=expanded_wires)
        return reduce_func(mat1, mat2), expanded_wires

    reduced_mat, final_wires = reduce(expand_and_reduce, mats_and_wires_gen)

    return reduced_mat, final_wires


def _local_sparse_swap_mat(i, n, format="csr"):
    """Helper function which generates the sparse matrix of SWAP
    for qubits: i <--> i+1 with final shape (2**n, 2**n)."""
    assert i < n - 1
    swap_mat = csr_matrix([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    j = (
        i + 1
    )  # i is the index of the qubit, j is the number of qubits prior to and include qubit i
    return kron(
        kron(eye(2 ** (j - 1)), swap_mat), eye(2 ** (n - (j + 1))), format=format
    )  # (j - 1) + 2 + (n - (j+1)) = n


def _sparse_swap_mat(i, j, n, format="csr"):
    """Helper function which generates the sparse matrix of SWAP
    for qubits: i <--> j with final shape (2**n, 2**n)."""
    assert i < n and j < n
    if i == j:
        return eye(2**n, format=format)

    (small_i, big_j) = (i, j) if i < j else (j, i)
    store_swaps = [
        _local_sparse_swap_mat(index, n, format=format)
        for index in range(small_i, big_j)
    ]

    res = eye(2**n, format=format)
    for mat in store_swaps:  # swap i --> j
        res @= mat

    for mat in store_swaps[-2::-1]:  # bring j --> old_i
        res @= mat

    return res


def _sparse_expand_matrix(base_matrix, wires, wire_order, format="csr"):
    """Re-express a sparse base matrix acting on a subspace defined by a set of wire labels
    according to a global wire order.

    Args:
        base_matrix (scipy.sparse.spmatrix): base matrix to expand
        wires (Iterable): wires determining the subspace that base matrix acts on; a base matrix of
            dimension :math:`2^n` acts on a subspace of :math:`n` wires
        wire_order (Iterable): global wire order, which has to contain all wire labels in ``wires``, but can also
            contain additional labels
        format (str): string representing the preferred scipy sparse matrix format to cast the expanded
            matrix too

    Returns:
        tensor_like: expanded matrix
    """
    n_wires = len(wires)
    n_total_wires = len(wire_order)

    if isinstance(wires, qml.wires.Wires):
        expanded_wires = wires.tolist()
    else:
        expanded_wires = list(copy.copy(wires))

    for wire in wire_order:
        if wire not in wires:
            expanded_wires.append(wire)

    num_missing_wires = n_total_wires - n_wires
    if num_missing_wires > 0:
        expanded_matrix = kron(
            base_matrix, eye(2**num_missing_wires, format=format), format=format
        )  # added missing wires at the end
    else:
        expanded_matrix = copy.copy(base_matrix)

    U = eye(2**n_total_wires, format=format)
    for i in range(n_total_wires):
        if expanded_wires[i] != wire_order[i]:
            j = expanded_wires.index(wire_order[i])  # location of correct wire
            U = U @ _sparse_swap_mat(
                i, j, n_total_wires, format=format
            )  # swap incorrect wire for correct wire

            expanded_wires[i], expanded_wires[j] = expanded_wires[j], expanded_wires[i]

    expanded_matrix = U.T @ expanded_matrix @ U
    return expanded_matrix.asformat(format)



"""
Test the plot functionalities in qaoa_parameters 
"""

# import the standard modules from python
import unittest

# import the OpenQAOA Parameterisation classes manually: Manual Mode
from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    PauliOp,
    Hamiltonian,
    QAOADescriptor,
)

# import the other OpenQAOA modules required for this example
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.problems import MaximumCut


# Create a hamiltonian problem on 15 qubits and create Circuit Params Class
hamiltonian1 = MaximumCut.random_instance(
    n_nodes=15, edge_probability=0.9
).qubo.hamiltonian
mixer_hamiltonian1 = X_mixer_hamiltonian(n_qubits=15)
qaoa_descriptor1 = QAOADescriptor(
    cost_hamiltonian=hamiltonian1, mixer_block=mixer_hamiltonian1, p=4
)

# Create a hamiltonian on 3 qubits with 2 coupling terms and 1 bias term and create Circuit Params Class
Term1 = PauliOp("ZZ", (0, 1))
Term2 = PauliOp("ZZ", (0, 2))
Term3 = PauliOp("Z", (0,))

hamiltonian2 = Hamiltonian([Term1, Term2, Term3], [0.7, 1.2, -0.5], 0.0)
mixer_hamiltonian2 = X_mixer_hamiltonian(n_qubits=3)
qaoa_descriptor2 = QAOADescriptor(
    cost_hamiltonian=hamiltonian2, mixer_block=mixer_hamiltonian2, p=4
)


class TestingQAOAParametersPlots(unittest.TestCase):
    def _plot(self, qaoa_descriptor, params_type, init_type="rand"):
        # generic function to create the plots
        print(f"Example of {params_type} params")
        if params_type in ["fourier", "fourier_extended", "fourier_w_bias"]:
            params = create_qaoa_variational_params(
                qaoa_descriptor=qaoa_descriptor,
                params_type=params_type,
                init_type=init_type,
                q=2,
            )
        else:
            params = create_qaoa_variational_params(
                qaoa_descriptor=qaoa_descriptor,
                params_type=params_type,
                init_type=init_type,
            )
        params.plot()

    def test_QAOAVariationalStandardParamsPlots(self):  # standard params
        self._plot(qaoa_descriptor1, "standard")

    def test_QAOAVariationalStandardBiasParamsPlots(self):  # standard_w_bias params
        self._plot(qaoa_descriptor2, "standard_w_bias")

    def test_QAOAVariationalExtendedParamsPlots(self):  # extended params
        self._plot(qaoa_descriptor1, "extended")
        self._plot(qaoa_descriptor2, "extended")

    def test_QAOAVariationalFourierParamsPlots(self):  # fourier params
        self._plot(qaoa_descriptor2, "fourier")

    def test_QAOAVariationalFourierBiasParamsPlots(self):  # fourier_w_bias params
        self._plot(qaoa_descriptor2, "fourier_w_bias")

    def test_QAOAVariationalFourierExtendedParamsPlots(self):  # fourier_extended params
        self._plot(qaoa_descriptor1, "fourier_extended")
        self._plot(qaoa_descriptor2, "fourier_extended")

    def test_QAOAVariationalAnnealingParamsPlots(self):  # annealing params
        self._plot(qaoa_descriptor2, "annealing", init_type="ramp")


if __name__ == "__main__":
    unittest.main()



import numpy as np

from typing import Union
from docplex.mp.model import Model

from .problem import Problem
from .converters import FromDocplex2IsingModel
from .qubo import QUBO


class BinPacking(Problem):
    """
    Creates an instance of the bin packing problem.
    https://en.wikipedia.org/wiki/Bin_packing_problem

    Parameters
    ----------
    weights: List[int]
        The weight of the items that must be placed in the bins.
    weight_capacity: int
        The maximum weight the bin can hold.
    penalty: float
        Penalty for the weight constraint.

    Returns
    -------
        An instance of the bin packing problem.
    """

    __name__ = "bin_packing"

    def __init__(
        self,
        weights: list = [],
        weight_capacity: int = 0,
        penalty: list = [],
        n_bins: int = None,
        simplifications=True,
        method: str = "slack",
    ):
        # include_ineqs: True if including the inequalities
        if method not in ["slack", "unbalanced"]:
            raise ValueError(
                f"The method '{method}' is not a valid method. Choose between 'slack' and 'unbalanced'"
            )
        for weight in weights:
            if int(weight) != weight:
                raise TypeError(
                    f"The weights must be integer numbers. Format {type(weight)} found."
                )

        if int(weight_capacity) != weight_capacity:
            raise TypeError(
                f"The weight_capacity must be integer. Format {type(weight_capacity)} found."
            )
        self.weights = [int(weight) for weight in weights]
        self.weight_capacity = weight_capacity
        self.penalty = penalty
        self.n_items = len(weights)
        self.method = method
        self.simplifications = simplifications
        if n_bins is None:
            self.n_bins = self.n_items
        else:
            self.n_bins = n_bins
        self.solution = self.solution_dict

    @property
    def solution_dict(self):
        """
        Dictionary with keys: variables and values: binary representation of the variables
        y_{j} -> represents if bin j is used
        x_{i}_{j} -> represent if item i is in bin j

        Returns
        -------
        solution : dict
        Dictionary with keys equal to the variables of the problem and values 0, 1, or None
        None for those values to be optimized otherwise the simplifications that can be
        used in this problem.

        """
        solution = {f"y_{j}": -1 for j in range(self.n_bins)}
        for i in range(self.n_items):
            for j in range(self.n_bins):
                solution[f"x_{i}_{j}"] = -1
        if self.simplifications:
            # First simplification: we know the minimum number of bins
            self.min_bins = int(np.ceil(np.sum(self.weights) / self.weight_capacity))
            for j in range(self.n_bins):
                if j < self.min_bins:
                    solution[f"y_{j}"] = 1
            solution["x_0_0"] = 1
            for j in range(1, self.n_bins):
                solution[f"x_0_{j}"] = 0
        return solution

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Bin Packing problem.

        Parameters
        ----------
        n_items: int
            The number of items that must be placed in the bins.
        seed: int
            Seed number for choosing the random weights
        min_weight: int
            Minimum weight to choose the weights from
        max_weight: int
            Maximum weight to choose the random weights from
        simplification: bool
            Use the simplification version of the problem
        method: ["slack", "unbalanced"]
            Use one of the two methods available to enconde the inequality constraints.
            The slack method requires extra-qubits.

        Returns
        -------
            A random instance of the bin packing problem.
        """
        seed = kwargs.get("seed", 1234)
        rng = np.random.default_rng(seed)
        weight_capacity = kwargs.get("weight_capacity", 15)
        n_items = kwargs.get("n_items", 3)
        n_bins = n_items
        min_weight = kwargs.get("min_weight", 1)
        max_weight = kwargs.get("max_weight", 7)
        if min_weight >= max_weight:
            raise ValueError(
                f"min_weight: {min_weight} must be < max_weight:{max_weight}"
            )
        weights = [rng.integers(min_weight, max_weight) for _ in range(n_items)]
        simplification = kwargs.get("simplification", True)
        penalty = kwargs.get("penalty", [])
        method = kwargs.get("method", "slack")
        return BinPacking(
            weights, weight_capacity, penalty, n_bins, simplification, method
        )

    @property
    def docplex_model(self):
        mdl = Model("bin_packing")
        vars_ = {}
        for var in self.solution.keys():
            if self.solution[var] == -1:
                vars_[var] = mdl.binary_var(var)
            else:
                vars_[var] = self.solution[var]
        objective = mdl.sum([vars_[y] for y in vars_.keys() if y[0] == "y"])
        vars_pos = {var.name: n for n, var in enumerate(mdl.iter_binary_vars())}

        mdl.minimize(objective)
        if self.simplifications:
            list_items = range(1, self.n_items)
        else:
            list_items = range(self.n_items)

        eq_constraints = {}
        ineq_constraints = {}
        for i in list_items:
            # First set of constraints: the items must be in any bin
            eq_constraints[f"eq_{i}"] = [
                [vars_pos[f"x_{i}_{j}"] for j in range(self.n_bins)],
                [1],
            ]
            mdl.add_constraint(
                mdl.sum(vars_[f"x_{i}_{j}"] for j in range(self.n_bins)) == 1
            )

        for j in range(self.n_bins):
            # Second set of constraints: weight constraints
            mdl.add_constraint(
                mdl.sum(
                    (self.weights[i] * vars_[f"x_{i}_{j}"] for i in range(self.n_items))
                )
                <= self.weight_capacity * vars_[f"y_{j}"]
            )
            if self.simplifications and j < self.min_bins:
                if j == 0:
                    ineq_constraints[f"ineq_{j}"] = [
                        [vars_pos[f"x_{i}_{j}"] for i in list_items],
                        [self.weight_capacity - self.weights[0]],
                    ]
                else:
                    ineq_constraints[f"ineq_{j}"] = [
                        [vars_pos[f"x_{i}_{j}"] for i in list_items],
                        [self.weight_capacity],
                    ]

            else:
                ineq_constraints[f"ineq_{j}"] = [
                    [vars_pos[f"x_{i}_{j}"] for i in list_items],
                    [self.weight_capacity * vars_pos[f"y_{j}"]],
                ]

        return mdl

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        cplex_model = self.docplex_model
        n_vars = cplex_model.number_of_binary_variables
        if len(self.penalty) > 0:
            if self.method == "slack":
                qubo_docplex = FromDocplex2IsingModel(
                    cplex_model, multipliers=self.penalty[0]
                )
            elif self.method == "unbalanced":
                qubo_docplex = FromDocplex2IsingModel(
                    cplex_model,
                    multipliers=self.penalty[0],
                    unbalanced_const=True,
                    strength_ineq=self.penalty[1:],
                )
        else:
            if self.method == "slack":
                qubo_docplex = FromDocplex2IsingModel(cplex_model)
            elif self.method == "unbalanced":
                qubo_docplex = FromDocplex2IsingModel(
                    cplex_model, unbalanced_const=True
                )

        ising_model = qubo_docplex.ising_model
        return QUBO(
            n_vars,
            ising_model.terms + [[]],
            ising_model.weights + [ising_model.constant],
            self.problem_instance,
        )

    def classical_solution(self, string: bool = False):
        """
        Return the classical solution of the bin packing problem

        Parameters
        ----------
        string : bool, optional
            If the solution is returned as a string. The default is False.

        Raises
        ------
        ValueError
            A flag if docplex does not find a valid solution.

        Returns
        -------
        solution : Union[str, dict]
            The classical solution of the specific problem as a string or a dict.

        """
        cplex_model = self.docplex_model
        docplex_sol = cplex_model.solve()

        if docplex_sol is None:
            raise ValueError(f"solution not found: {cplex_model.solve_details.status}")

        if string:
            solution = ""
        else:
            solution = self.solution.copy()
        for var in cplex_model.iter_binary_vars():
            if string:
                solution += str(int(np.round(docplex_sol.get_value(var), 1)))
            else:
                solution[var.name] = int(np.round(docplex_sol.get_value(var), 1))
        return solution

    def plot_solution(self, solution: Union[dict, str], ax=None):
        """
        A visualization method for the bin packing problem solution.

        Parameters
        ----------
        solution : Union[dict, str]
            The solution of the specific bin packing problem as a string or dictionary.
        ax : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        fig : matplotlib.pyplot.Figure()
            The plot visualization of the solution.

        """
        import matplotlib.pyplot as plt
        from matplotlib import colormaps

        cplex_model = self.docplex_model
        if isinstance(solution, str):
            sol = self.solution.copy()
            for n, var in enumerate(cplex_model.iter_binary_vars()):
                sol[var.name] = int(solution[n])
            solution = sol
        colors = colormaps["jet"]
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        for j in range(self.n_bins):
            sum_items = 0
            if solution[f"y_{j}"]:
                for i in range(self.n_items):
                    if solution[f"x_{i}_{j}"]:
                        ax.bar(
                            j,
                            self.weights[i],
                            bottom=sum_items,
                            label=f"item {i}",
                            color=colors(i / self.n_items),
                            alpha=0.7,
                            edgecolor="black",
                        )
                        sum_items += self.weights[i]
        ax.hlines(
            self.weight_capacity,
            -0.5,
            self.n_bins - 0.5,
            linestyle="--",
            color="black",
            label="Max Weight",
        )
        ax.set_xticks(np.arange(self.n_bins), np.arange(self.n_bins), fontsize=14)
        ax.set_xlabel("bin", fontsize=14)
        ax.set_ylabel("weight", fontsize=14)
        ax.legend(
            loc="upper center",
            bbox_to_anchor=(0.5, 1.2 + 0.011 * self.n_items),
            ncol=5,
            fancybox=True,
            shadow=True,
        )
        return fig



import copy
import numpy as np
from typing import List
import matplotlib.pyplot as plt

from ...utilities import delete_keys_from_dict
from .. import QAOAResult
from ...problems import QUBO


class RQAOAResult(dict):
    """
    A class to handle the results of RQAOA workflows
    It stores the results of the RQAOA optimization as a dictionary. With some custom methods.
    """

    def asdict(
        self,
        keep_cost_hamiltonian: bool = True,
        complex_to_string: bool = False,
        intermediate_measurements: bool = True,
        exclude_keys: List[str] = [],
    ):
        """
        Returns the results as a full dictionary, meaning that the objects of the
        intermediate steps are also converted to dictionaries.

        Parameters
        ----------
        keep_cost_hamiltonian : bool, optional
            If True, the cost Hamiltonian is kept in the dictionary, by default True.
        complex_to_string : bool, optional
            If True, the complex numbers are converted to strings, by default False.
            This is useful for JSON serialization.
        intermediate_measurements: bool, optional
            If True, intermediate measurements are included in the dump. If False,
            intermediate measurements are not included in the dump.
            Default is True.
        exclude_keys: `list[str]`, optional
            A list of keys to exclude from the returned dictionary.

        Returns
        -------
        dict
            The results as a dictionary.
        """

        results = {k: v for k, v in self.items()}
        results["intermediate_steps"] = []
        for step in self["intermediate_steps"]:
            results["intermediate_steps"].append(
                {
                    "counter": step["counter"],
                    "problem": step["problem"].asdict(),
                    "qaoa_results": step["qaoa_results"].asdict(
                        keep_cost_hamiltonian,
                        complex_to_string,
                        intermediate_measurements,
                    ),
                    "exp_vals_z": step["exp_vals_z"].tolist(),
                    "corr_matrix": step["corr_matrix"].tolist(),
                }
            )
        return (
            results
            if exclude_keys == []
            else delete_keys_from_dict(results, exclude_keys)
        )

    @classmethod
    def from_dict(cls, dictionary: dict):
        """
        Creates a RQAOAResult object from a dictionary (which is the output of the asdict method).
        Parameters
        ----------
        dictionary : dict
            The input dictionary.
        Returns
        -------
        RQAOAResult
            The RQAOAResult object.
        """

        # deepcopy the dictionary, so that the original dictionary is not changed
        dictionary = copy.deepcopy(dictionary)

        # create a new RQAOAResult object
        results = cls()

        # add the keys of the dictionary to the RQAOAResult object
        for key, value in dictionary.items():
            results[key] = value

        # convert the intermediate steps to objects
        for step in results["intermediate_steps"]:
            step["problem"] = QUBO.from_dict(step["problem"])
            step["qaoa_results"] = QAOAResult.from_dict(
                step["qaoa_results"], cost_hamiltonian=step["problem"].hamiltonian
            )
            step["exp_vals_z"] = np.array(step["exp_vals_z"])
            step["corr_matrix"] = np.array(step["corr_matrix"])

        return results

    def get_solution(self):
        """
        Returns the solution of the optimization.
        """
        return self["solution"]

    def get_qaoa_results(self, step):
        """
        Returns the i-th qaoa step of the RQAOA.
        """
        return self["intermediate_steps"][step]["qaoa_results"]

    def get_qaoa_optimized_angles(self, step):
        """
        Returns the optimized angles of the i-th qaoa step of the RQAOA.
        """
        return self.get_qaoa_results(step).optimized["angles"]

    def get_problem(self, step):
        """
        Returns the QUBO problem in the i-th step of the RQAOA.
        """
        return self["intermediate_steps"][step]["problem"]

    def get_hamiltonian(self, step):
        """
        Returns the Hamiltonian of the i-th step of the RQAOA.
        """
        return self.get_problem(step).hamiltonian

    def get_exp_vals_z(self, step):
        """
        Returns the expectation values of the Z operator of the i-th step of the RQAOA.
        """
        return self["intermediate_steps"][step]["exp_vals_z"]

    def get_corr_matrix(self, step):
        """
        Returns the correlation matrix of the i-th step of the RQAOA.
        """
        return self["intermediate_steps"][step]["corr_matrix"]

    def plot_corr_matrix(self, step, cmap="cool"):
        """
        Plots the correlation matrix of the i-th step of the RQAOA.
        TODO : add more options
        """
        plt.imshow(self.get_corr_matrix(step=step), cmap=cmap)
        plt.colorbar()



"""
Test and validate the creation of variational parameters and circuit parameters. 
"""
import unittest
import numpy as np
from scipy.fft import dst, dct

from openqaoa.qaoa_components import *
from openqaoa.utilities import X_mixer_hamiltonian, XY_mixer_hamiltonian
from openqaoa.qaoa_components.ansatz_constructor.gatemaplabel import GateMapType
from openqaoa.qaoa_components.variational_parameters.variational_params_factory import (
    PARAMS_CLASSES_MAPPER,
    VARIATIONAL_PARAMS_DICT_KEYS,
)

register = [0, 1, 2]
terms = [[0, 1], [2], [0, 2]]
weights = [1, 0.5, -2.0]

terms_wo_bias = [[0, 1], [1, 2]]
weights_wo_bias = [1, 0.5]

cost_hamiltonian = Hamiltonian.classical_hamiltonian(terms, weights, constant=0)
cost_hamiltonian_wo_bias = Hamiltonian.classical_hamiltonian(
    terms_wo_bias, weights_wo_bias, constant=0.7
)
mixer_hamiltonian = X_mixer_hamiltonian(len(register))


class TestingQAOADescriptor(unittest.TestCase):
    def setUp(self):
        self.p = 2
        self.cost_hamil = Hamiltonian(
            [
                PauliOp("ZZ", (0, 1)),
                PauliOp("ZZ", (1, 2)),
                PauliOp("ZZ", (0, 2)),
                PauliOp("Z", (0,)),
            ],
            [1, 1, 1, 0.5],
            1,
        )
        self.mixer_hamil = X_mixer_hamiltonian(n_qubits=3)
        self.mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2)]
        self.mixer_gatemap_coeffs = self.mixer_hamil.coeffs

    def test_QAOADescriptor(self):
        """
        QAOADescriptor accept 2 types of inputs for mixer_blocks argument on initilisation. This test checks that the same mixer block when presented in Hamiltonian or as a List of RotationGateMap, plus the appropriate mixer coefficients, produces similar internal attributes.
        """

        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p=self.p)

        qaoa_descriptor_gm = QAOADescriptor(
            self.cost_hamil,
            self.mixer_gatemap,
            p=self.p,
            mixer_coeffs=self.mixer_gatemap_coeffs,
        )

        # Check same number of mixer gates
        self.assertEqual(len(qaoa_descriptor.mixer_blocks), self.p)
        self.assertEqual(len(qaoa_descriptor_gm.mixer_blocks), self.p)
        for each_p_value in range(self.p):
            self.assertEqual(len(qaoa_descriptor.mixer_blocks[each_p_value]), 3)
            self.assertEqual(len(qaoa_descriptor_gm.mixer_blocks[each_p_value]), 3)
        # Check similar coefficients and qubit singles/pairs names
        self.assertEqual(
            qaoa_descriptor.mixer_single_qubit_coeffs,
            qaoa_descriptor_gm.mixer_single_qubit_coeffs,
        )
        self.assertEqual(
            qaoa_descriptor.mixer_pair_qubit_coeffs,
            qaoa_descriptor_gm.mixer_pair_qubit_coeffs,
        )
        self.assertEqual(
            qaoa_descriptor.mixer_qubits_singles,
            qaoa_descriptor_gm.mixer_qubits_singles,
        )
        self.assertEqual(
            qaoa_descriptor.mixer_qubits_pairs, qaoa_descriptor_gm.mixer_qubits_pairs
        )

    def test_QAOADescriptor_mixer_coeffs_selector_hamiltonian(self):
        """
        If a mixer hamiltonian is used for mixer_block, the coefficients will be obtained from it. Even if the user inputs his own mixer coefficient, it will be ignored.
        """

        verify_mixer_coeffs = [-1, -1, -1]

        mixer_hamil = X_mixer_hamiltonian(n_qubits=3)

        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p=self.p)
        qaoa_descriptor_2 = QAOADescriptor(
            self.cost_hamil, self.mixer_hamil, p=self.p, mixer_coeffs=[1, 0, 1]
        )

        self.assertEqual(qaoa_descriptor.mixer_block_coeffs, verify_mixer_coeffs)
        self.assertEqual(qaoa_descriptor_2.mixer_block_coeffs, verify_mixer_coeffs)

    def test_QAOADescriptor_mixer_coeffs_selector_gatemap(self):
        """
        If a mixer gatemap is used for mixer_block, the user is required to input his own mixer coefficients, there should be an error message if the user does not input the appropriate number of mixer coefficients.
        """

        verify_mixer_coeffs = [-1, -1, -1]

        qaoa_descriptor = QAOADescriptor(
            self.cost_hamil,
            self.mixer_gatemap,
            p=self.p,
            mixer_coeffs=self.mixer_gatemap_coeffs,
        )

        self.assertEqual(qaoa_descriptor.mixer_block_coeffs, verify_mixer_coeffs)

        with self.assertRaises(ValueError) as cm:
            QAOADescriptor(self.cost_hamil, self.mixer_gatemap, self.p, [])
            self.assertEqual(
                "The number of terms/gatemaps must match the number of coefficients provided.",
                str(cm.exception),
            )

    def test_QAOADescriptor_assign_coefficients(self):
        """
        The method should split the coefficients and gatemaps in the proper order. Regardless of their positions within the hamiltonian or gatemap list.
        """

        verify_mixer_coeffs = [-1, -1, -1, -0.5]

        # TestCase 1: GateMap Mixer
        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2), RXXGateMap(0, 2)]

        qaoa_descriptor = QAOADescriptor(
            self.cost_hamil, mixer_gatemap, p=self.p, mixer_coeffs=[-1, -1, -1, -0.5]
        )

        self.assertEqual(qaoa_descriptor.cost_single_qubit_coeffs, [0.5])
        self.assertEqual(qaoa_descriptor.cost_pair_qubit_coeffs, [1, 1, 1])
        self.assertEqual(qaoa_descriptor.cost_qubits_singles, ["RZGateMap"])
        self.assertEqual(
            qaoa_descriptor.cost_qubits_pairs,
            ["RZZGateMap", "RZZGateMap", "RZZGateMap"],
        )

        self.assertEqual(qaoa_descriptor.mixer_single_qubit_coeffs, [-1, -1, -1])
        self.assertEqual(qaoa_descriptor.mixer_pair_qubit_coeffs, [-0.5])
        self.assertEqual(
            qaoa_descriptor.mixer_qubits_singles,
            ["RXGateMap", "RXGateMap", "RXGateMap"],
        )
        self.assertEqual(qaoa_descriptor.mixer_qubits_pairs, ["RXXGateMap"])

        # TestCase 2: Hamiltonian Mixer
        mixer_gatemap = XY_mixer_hamiltonian(n_qubits=3)

        qaoa_descriptor = QAOADescriptor(self.cost_hamil, mixer_gatemap, p=self.p)

        self.assertEqual(qaoa_descriptor.cost_single_qubit_coeffs, [0.5])
        self.assertEqual(qaoa_descriptor.cost_pair_qubit_coeffs, [1, 1, 1])
        self.assertEqual(qaoa_descriptor.cost_qubits_singles, ["RZGateMap"])
        self.assertEqual(
            qaoa_descriptor.cost_qubits_pairs,
            ["RZZGateMap", "RZZGateMap", "RZZGateMap"],
        )

        self.assertEqual(qaoa_descriptor.mixer_single_qubit_coeffs, [])
        self.assertEqual(
            qaoa_descriptor.mixer_pair_qubit_coeffs, [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
        )
        self.assertEqual(qaoa_descriptor.mixer_qubits_singles, [])
        self.assertEqual(
            qaoa_descriptor.mixer_qubits_pairs,
            [
                "RXXGateMap",
                "RYYGateMap",
                "RXXGateMap",
                "RYYGateMap",
                "RXXGateMap",
                "RYYGateMap",
            ],
        )

    def test_QAOADescriptor_cost_blocks(self):
        """
        cost_blocks property should always return a list of RotationGateMaps based on the input cost hamiltonian.
        """

        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2), RXXGateMap(0, 2)]

        qaoa_descriptor = QAOADescriptor(
            self.cost_hamil, mixer_gatemap, p=self.p, mixer_coeffs=[-1, -1, -1, -0.5]
        )

        for p_index, each_p_block in enumerate(qaoa_descriptor.cost_blocks):
            for each_item in each_p_block:
                self.assertTrue(isinstance(each_item, RotationGateMap))
                self.assertEqual(each_item.gate_label.type, GateMapType.COST)
                self.assertEqual(each_item.gate_label.layer, p_index)

            self.assertEqual(
                [each_item.gate_label.sequence for each_item in each_p_block],
                [0, 1, 2, 0],
            )

    def test_QAOADescriptor_mixer_blocks(self):
        """
        mixer_blocks property should always return a list of RotationGateMaps based on the input cost hamiltonian.
        """

        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2), RXXGateMap(0, 2)]

        qaoa_descriptor = QAOADescriptor(
            self.cost_hamil, mixer_gatemap, p=self.p, mixer_coeffs=[-1, -1, -1, -0.5]
        )

        for p_index, each_p_block in enumerate(qaoa_descriptor.mixer_blocks):
            for each_item in each_p_block:
                self.assertTrue(isinstance(each_item, RotationGateMap))
                self.assertEqual(each_item.gate_label.type, GateMapType.MIXER)
                self.assertEqual(each_item.gate_label.layer, p_index)

            self.assertEqual(
                [each_item.gate_label.sequence for each_item in each_p_block],
                [0, 1, 2, 0],
            )

    def test_QAOADescriptor_weird_cases(self):
        """
        Testing various weird cases
        """

        # TestCase 1: Empty List Mixer Block
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, [], p=self.p)

        # TestCase 2: Wrong Type Cost Block
        self.assertRaises(AttributeError, QAOADescriptor, [], self.mixer_hamil, self.p)
        self.assertRaises(
            AttributeError, QAOADescriptor, [RZGateMap(0)], self.mixer_hamil, self.p
        )

        # TestCase 3: p is not an int
        self.assertRaises(
            TypeError, QAOADescriptor, self.cost_hamil, self.mixer_hamil, "test"
        )

    def test_QAOADescriptor_abstract_circuit(self):
        """
        The abstract circuit should be consistent with the cost hamiltonian, mixer block and the p value.
        """

        mixer_hamil = X_mixer_hamiltonian(n_qubits=3)
        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2)]

        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p=self.p)
        qaoa_descriptor_gm = QAOADescriptor(
            self.cost_hamil,
            self.mixer_gatemap,
            p=self.p,
            mixer_coeffs=self.mixer_gatemap_coeffs,
        )

        # Checks if the gates are identical
        for each_cp_gate, each_cpgm_gate in zip(
            qaoa_descriptor.abstract_circuit, qaoa_descriptor_gm.abstract_circuit
        ):
            self.assertTrue(type(each_cp_gate), type(each_cpgm_gate))
            if hasattr(each_cp_gate, "qubit_1"):
                self.assertEqual(each_cp_gate.qubit_1, each_cpgm_gate.qubit_1)
            if hasattr(each_cpgm_gate, "qubit_2"):
                self.assertEqual(each_cp_gate.qubit_2, each_cpgm_gate.qubit_2)


class TestingQAOAVariationalParameters(unittest.TestCase):
    def test_QAOAVariationalExtendedParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2
        )

        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.betas_singles, [[0.75] * 3, [0.25] * 3])
        assert np.allclose(params.betas_pairs, [])
        assert np.allclose(params.gammas_singles, [[0.25], [0.75]])
        assert np.allclose(params.gammas_pairs, [[0.25, 0.25], [0.75, 0.75]])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    # TODO check that the values also make sense
    def test_QAOAVariationalExtendedParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)

        betas_singles = [[0.0, 0.1, 0.3], [0.5, 0.2, 1.2]]
        betas_pairs = []
        gammas_singles = [[0.0], [0.5]]
        gammas_pairs = [[0.1, 0.3], [0.2, 1.2]]
        params = QAOAVariationalExtendedParams(
            qaoa_descriptor, betas_singles, betas_pairs, gammas_singles, gammas_pairs
        )

        cost_2q_angles = 2 * np.array([1, -2.0]) * gammas_pairs
        cost_1q_angles = 2 * np.array([0.5]) * gammas_singles
        mixer_1q_angles = 2 * np.array([-1, -1, -1]) * betas_singles
        mixer_2q_angles = []

        assert np.allclose(params.betas_singles, betas_singles)
        assert np.allclose(params.betas_pairs, betas_pairs)
        assert np.allclose(params.gammas_singles, gammas_singles)
        assert np.allclose(params.gammas_pairs, gammas_pairs)

        assert np.allclose(params.mixer_1q_angles, mixer_1q_angles)
        assert np.allclose(params.mixer_2q_angles, mixer_2q_angles)
        assert np.allclose(params.cost_1q_angles, cost_1q_angles)
        assert np.allclose(params.cost_2q_angles, cost_2q_angles)

    def test_QAOAVariationalStandardWithBiasParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)

        betas = [np.pi, 0.4]
        gammas_singles = [10, 24]
        gammas_pairs = [8.8, 2.3]
        params = QAOAVariationalStandardWithBiasParams(
            qaoa_descriptor, betas, gammas_singles, gammas_pairs
        )

        cost_2q_angles = 2 * np.outer(gammas_pairs, np.array([1, -2.0]))
        cost_1q_angles = 2 * np.outer(gammas_singles, np.array([0.5]))
        mixer_1q_angles = 2 * np.outer(betas, np.array([-1, -1, -1]))
        mixer_2q_angles = []

        assert np.allclose(params.betas, betas)
        assert np.allclose(params.gammas_singles, gammas_singles)
        assert np.allclose(params.gammas_pairs, gammas_pairs)

        assert np.allclose(params.mixer_1q_angles, mixer_1q_angles)
        assert np.allclose(params.mixer_2q_angles, mixer_2q_angles)
        assert np.allclose(params.cost_1q_angles, cost_1q_angles)
        assert np.allclose(params.cost_2q_angles, cost_2q_angles)
        assert type(params) == QAOAVariationalStandardWithBiasParams

    def test_QAOAVariationalAnnealingParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        schedule = [0.4, 1.0]
        annealing_params = QAOAVariationalAnnealingParams(
            qaoa_descriptor, total_annealing_time=5, schedule=schedule
        )

        assert type(annealing_params) == QAOAVariationalAnnealingParams

    def test_QAOAVariationalFourierParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        v = [0.4, 1.0]
        u = [0.5, 1.2]
        fourier_params = QAOAVariationalFourierParams(qaoa_descriptor, q=2, v=v, u=u)

        assert np.allclose(fourier_params.betas, dct(v, n=qaoa_descriptor.p, axis=0))
        assert np.allclose(fourier_params.gammas, dst(u, n=qaoa_descriptor.p, axis=0))
        assert type(fourier_params) == QAOAVariationalFourierParams

    def test_QAOAVariationalFourierExtendedParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        v_singles = [[0.4] * 3, [1.0] * 3]
        v_pairs = []
        u_singles = [[0.5], [1.2]]
        u_pairs = [[4.5] * 2, [123] * 2]
        fourier_params = QAOAVariationalFourierExtendedParams(
            qaoa_descriptor, 2, v_singles, v_pairs, u_singles, u_pairs
        )

        assert np.allclose(
            fourier_params.betas_singles, dct(v_singles, n=qaoa_descriptor.p, axis=0)
        )
        assert np.allclose(
            fourier_params.gammas_singles, dst(u_singles, n=qaoa_descriptor.p, axis=0)
        )
        assert np.allclose(
            fourier_params.gammas_pairs, dst(u_pairs, n=qaoa_descriptor.p, axis=0)
        )

        assert type(fourier_params) == QAOAVariationalFourierExtendedParams

    def test_QAOAVariationalAnnealingParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalAnnealingParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, total_annealing_time=2
        )

        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.mixer_1q_angles, [[-2 * 0.75] * 3, [-2 * 0.25] * 3])
        assert np.allclose(params.cost_1q_angles, [[2 * 0.125], [2 * 0.375]])
        assert np.allclose(
            params.cost_2q_angles, [[2 * 0.25, -0.5 * 2], [2 * 0.75, -1.5 * 2]]
        )

        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalStandardWithBiasParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalStandardWithBiasParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2
        )
        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.mixer_1q_angles, [[-2 * 0.75] * 3, [-2 * 0.25] * 3])
        assert np.allclose(params.cost_1q_angles, [[2 * 0.125], [2 * 0.375]])
        assert np.allclose(
            params.cost_2q_angles, [[2 * 0.25, -0.5 * 2], [2 * 0.75, -1.5 * 2]]
        )

        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalStandardParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalStandardParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2
        )

        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.mixer_1q_angles, [[-2 * 0.75] * 3, [-2 * 0.25] * 3])
        assert np.allclose(params.cost_1q_angles, [[2 * 0.125], [2 * 0.375]])
        assert np.allclose(
            params.cost_2q_angles, [[2 * 0.25, -0.5 * 2], [2 * 0.75, -1.5 * 2]]
        )

        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_non_fourier_params_are_consistent(self):
        """
        Check that StandardParams, StandardWithBiasParams and
        ExtendedParams give the same rotation angles, given the same data"""
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)
        p1 = QAOAVariationalStandardParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2
        )
        p2 = QAOAVariationalExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2
        )
        p3 = QAOAVariationalStandardWithBiasParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2
        )

        assert np.allclose(p1.mixer_1q_angles, p2.mixer_1q_angles)
        assert np.allclose(p2.mixer_1q_angles, p3.mixer_1q_angles)
        assert np.allclose(p1.cost_1q_angles, p2.cost_1q_angles)
        assert np.allclose(p2.cost_1q_angles, p3.cost_1q_angles)
        assert np.allclose(p1.cost_2q_angles, p2.cost_2q_angles)
        assert np.allclose(p2.cost_2q_angles, p3.cost_2q_angles)

    def test_QAOAVariationalFourierParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=3)
        params = QAOAVariationalFourierParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2
        )

        # just access the angles, to check that it actually creates them
        assert len(params.cost_1q_angles) == len(params.cost_2q_angles)
        assert np.allclose(params.v, [1 / 3, 0])
        assert np.allclose(params.u, [1 / 3, 0])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalFourierWithBiasParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=3)
        params = QAOAVariationalFourierWithBiasParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2
        )

        # just access the angles, to check that it actually creates them
        assert len(params.cost_1q_angles) == len(params.cost_2q_angles)
        assert np.allclose(params.v, [1 / 3, 0])
        assert np.allclose(params.u_singles, [1 / 3, 0])
        assert np.allclose(params.u_pairs, [1 / 3, 0])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalFourierExtendedParams(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=3)
        params = QAOAVariationalFourierExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2
        )

        # just access the angles, to check that it actually creates them
        assert len(params.cost_1q_angles) == len(params.cost_2q_angles)
        # Test updating and raw output

        assert np.allclose(params.v_singles, [[1 / 3] * 3, [0] * 3])
        assert np.allclose(params.u_singles, [[1 / 3] * 1, [0] * 1])
        assert np.allclose(params.u_pairs, [[1 / 3] * 2, [0] * 2])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_FourierParams_are_consistent(self):
        """
        Check, that both Fourier Parametrizations give the same rotation angles,
        given the same data"""
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=3)

        params1 = QAOAVariationalFourierParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2
        )
        params2 = QAOAVariationalFourierWithBiasParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2
        )
        params3 = QAOAVariationalFourierExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2
        )

        assert np.allclose(params1.mixer_1q_angles, params2.mixer_1q_angles)
        assert np.allclose(params1.cost_1q_angles, params2.cost_1q_angles)
        assert np.allclose(params1.cost_2q_angles, params2.cost_2q_angles)
        assert np.allclose(params1.mixer_1q_angles, params3.mixer_1q_angles)
        assert np.allclose(params1.cost_1q_angles, params3.cost_1q_angles)
        assert np.allclose(params1.cost_2q_angles, params3.cost_2q_angles)

    def test_inputChecking(self):
        # Check that an error is raised if we pass in an extra angle in `betas` (depth is 3, here we give 4 beta values)
        reg = [0, 1]
        terms = [[0, 1]]
        weights = [0.7]
        n_layers = 3
        betas_singles = [1, 2, 3, 4]
        betas_pairs = []
        gammas_singles = []
        gammas_pairs = [1, 2, 3]
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=n_layers
        )
        self.assertRaises(
            ValueError,
            QAOAVariationalExtendedParams,
            qaoa_descriptor,
            betas_singles,
            betas_pairs,
            gammas_singles,
            gammas_pairs,
        )

    def test_str_and_repr(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)

        for each_key_value in PARAMS_CLASSES_MAPPER.keys():
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor,
                params_type=each_key_value,
                init_type="rand",
                q=1,
                total_annealing_time=1,
            )

            self.assertEqual(variate_params.__str__(), variate_params.__repr__())

    def test_qaoa_variational_params_asdict(self):
        p = 4
        q = 2

        number_of_params = {
            "standard": {"betas": p, "gammas": p},
            "standard_w_bias": {"betas": p, "gammas_singles": p, "gammas_pairs": p},
            "extended": {
                "betas_singles": len(register) * p,
                "betas_pairs": 0,
                "gammas_singles": sum([1 for term in terms if len(term) == 1]) * p,
                "gammas_pairs": sum([1 for term in terms if len(term) == 2]) * p,
            },
            "fourier": {"v": q, "u": q},
            "fourier_extended": {
                "v_singles": len(register) * q,
                "v_pairs": 0,
                "u_singles": sum([1 for term in terms if len(term) == 1]) * q,
                "u_pairs": sum([1 for term in terms if len(term) == 2]) * q,
            },
            "fourier_w_bias": {"v": q, "u_singles": q, "u_pairs": q},
            "annealing": {"schedule": p},
        }

        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=p)

        for each_key_value in PARAMS_CLASSES_MAPPER.keys():
            # create the variational params object
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor,
                params_type=each_key_value,
                init_type="rand",
                q=q,
                total_annealing_time=2,
            )

            # get the dict and check that it is a dict
            variate_params_dict = variate_params.asdict()
            assert isinstance(
                variate_params_dict, dict
            ), f"asdict() should return a dict, it did not for type: '{each_key_value}'."

            # check that the keys are correct
            expected_keys = VARIATIONAL_PARAMS_DICT_KEYS[each_key_value]
            for key in [
                "q",
                "total_annealing_time",
            ]:  # from expected_keys, remove the "q" and "total_annealing_time" keys, if there are any
                if key in expected_keys:
                    expected_keys.remove(key)
            assert set(variate_params_dict.keys()) == set(
                expected_keys
            ), f"asdict() should return a dict with the correct keys, it did not for type: '{each_key_value}'."

            # check that the number of values is correct
            for key, value in variate_params_dict.items():
                assert (
                    value.size == number_of_params[each_key_value][key]
                ), f"asdict() should return a dict with the correct number of values, it did not for type: '{each_key_value}'. \
                \n Expected {number_of_params[each_key_value][key]} values for key '{key}', but got {value.size} values."

            # check that the values are correct
            list_params_raw = []
            for key, value in variate_params_dict.items():
                list_params_raw += value.flatten().tolist()

            assert np.allclose(
                list_params_raw, variate_params.raw().tolist()
            ), f"asdict() should return a dict with the correct values, it did not for type: '{each_key_value}'."

    def test_qaoa_variational_params_update_from_dict(self):
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p=2)

        for each_key_value in PARAMS_CLASSES_MAPPER.keys():
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor,
                params_type=each_key_value,
                init_type="rand",
                q=1,
                total_annealing_time=1,
            )

            # get the params as a dict
            variate_params_dict = variate_params.asdict()

            # create a new dict with random values
            new_dict = {
                k: np.random.rand(*v.shape) for k, v in variate_params_dict.items()
            }

            # update the variational params object with the new dict
            variate_params.update_from_dict(new_dict)

            # check that the values are correct
            for key, value in variate_params.asdict().items():
                assert np.allclose(
                    value, new_dict[key]
                ), f"update_from_dict() should update the values correctly, it did not for type: '{each_key_value}'."

            # check that if we pass in a dict with the wrong keys, an error is raised
            wrong_dict = new_dict
            wrong_dict["wrong_key"] = np.random.rand(1)
            error = False
            try:
                variate_params.update_from_dict(wrong_dict)
            except:
                error = True
            assert (
                error
            ), f"update_from_dict() should raise an error if the dict has the wrong keys, it did not for type: '{each_key_value}'."

            # check that if we pass in a dict with more values than expected, an error is raised
            wrong_dict = new_dict
            wrong_dict.pop("wrong_key")
            for key in wrong_dict.keys():
                wrong_dict[key] = np.append(wrong_dict[key], np.random.rand(1))
            error = False
            try:
                variate_params.update_from_dict(wrong_dict)
            except Exception:
                error = True
            assert (
                error
            ), f"update_from_dict() should raise an error if the dict has more values than expected, it did not for type: '{each_key_value}'."

            # check that if we pass in a dict with less values than expected, an error is raised
            wrong_dict = new_dict
            for key in wrong_dict.keys():
                wrong_dict[key] = wrong_dict[key][:-2]
            error = False
            try:
                variate_params.update_from_dict(wrong_dict)
            except Exception:
                error = True
            assert (
                error
            ), f"update_from_dict() should raise an error if the dict has less values than expected, it did not for type: '{each_key_value}'."

            # check that if we pass something that is not a dict, an error is raised
            error = False
            try:
                variate_params.update_from_dict(1)
            except Exception:
                error = True
            assert (
                error
            ), f"update_from_dict() should raise an error if the input is not a dict, it did not for type: '{each_key_value}'."


if __name__ == "__main__":
    unittest.main()



"""
Optimizer directory for the classical optimization loop for QAOA 

Currently supports:
	ScipyOptimizers (both gradient-free and gradient-based)
	PennylaneOptimizers (adagrad, adam, gradient descent, nestrov momentum, rms prop, rotosolve, spsa)
"""

from .training_vqa import *
from .qaoa_optimizer import *



"""
Contains backend specific codes for implementing QAOA on supported simulators
and QPU backends

Currently Supported:
	Qiskit:
		QASM Simulator
		Statevector Simulator
		IBMQ available QPUs
	PyQuil:
		Rigetti QPUs
		Statevector Simulator
	Vectorized:
		Fast numpy native Statevector Simulator
"""
from .qaoa_vectorized import QAOAvectorizedBackendSimulator
from .qaoa_analytical_sim import QAOABackendAnalyticalSimulator
from .devices_core import DeviceLocal
from .qaoa_device import create_device



import unittest

import networkx as nx

from openqaoa.qaoa_components.ansatz_constructor import (
    GateMap,
    SWAPGateMap,
    RotationGateMap,
    RXGateMap,
    RZXGateMap,
    RXXGateMap,
    RYYGateMap,
)
from openqaoa.qaoa_components import QAOADescriptor, create_qaoa_variational_params
from openqaoa.backends import create_device
from openqaoa.optimizers import get_optimizer
from openqaoa.backends.qaoa_backend import get_qaoa_backend
from openqaoa.utilities import (
    quick_create_mixer_for_topology,
    X_mixer_hamiltonian,
    XY_mixer_hamiltonian,
)
from openqaoa.problems import MinimumVertexCover
from openqaoa.qaoa_components.ansatz_constructor.gatemaplabel import GateMapType


class TestingCustomMixer(unittest.TestCase):
    def setUp(self):
        nodes = 6
        edge_probability = 0.7
        g = nx.generators.fast_gnp_random_graph(n=nodes, p=edge_probability, seed=34)
        mini_cov = MinimumVertexCover(g, field=1.0, penalty=1.0)
        self.PROBLEM_QUBO = mini_cov.qubo

        # Case with Mixer with 1 AND 2-qubit terms
        self.MANUAL_GATEMAP_LIST, self.MANUAL_COEFFS = [
            RXGateMap(0),
            RZXGateMap(0, 1),
            RZXGateMap(0, 2),
            RZXGateMap(0, 3),
            RZXGateMap(0, 4),
            RXGateMap(4),
            RZXGateMap(0, 5),
            RXXGateMap(1, 2),
        ], [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        self.MANUAL_SEQUENCE = [0, 0, 1, 2, 3, 1, 4, 5]

        zx_gatemap_list, zx_gatemap_coeffs = quick_create_mixer_for_topology(
            RZXGateMap, 6, qubit_connectivity="star"
        )
        xx_gatemap_list, xx_gatemap_coeffs = quick_create_mixer_for_topology(
            RXXGateMap, 6, qubit_connectivity="full"
        )

        zx_gatemap_list.extend(xx_gatemap_list)
        zx_gatemap_coeffs.extend(xx_gatemap_coeffs)

        # Case with Multiple types of 2-qubit gates
        self.COMPLICATED_GATEMAP_LIST, self.COMPLICATED_COEFFS = (
            zx_gatemap_list,
            zx_gatemap_coeffs,
        )
        self.COMPLICATED_SEQUENCE = [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
        ]

        self.TESTING_GATEMAPS = [
            [self.MANUAL_GATEMAP_LIST, self.MANUAL_COEFFS, self.MANUAL_SEQUENCE],
            [
                self.COMPLICATED_GATEMAP_LIST,
                self.COMPLICATED_COEFFS,
                self.COMPLICATED_SEQUENCE,
            ],
        ]

    def test_custom_mixer_basic_workflow(self):
        """
        Check that using custom mixers works.
        Custom Mixers are only available in Manual mode.
        """

        for each_gatemap_list, each_gatemap_coeffs, _ in self.TESTING_GATEMAPS:
            custom_mixer_block_gatemap = each_gatemap_list
            custom_mixer_block_coeffs = each_gatemap_coeffs

            qaoa_descriptor = QAOADescriptor(
                self.PROBLEM_QUBO.hamiltonian,
                custom_mixer_block_gatemap,
                p=1,
                mixer_coeffs=custom_mixer_block_coeffs,
            )
            device_local = create_device(location="local", name="vectorized")
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "rand"
            )
            backend_local = get_qaoa_backend(qaoa_descriptor, device_local, n_shots=500)
            optimizer = get_optimizer(
                backend_local, variate_params, {"method": "cobyla", "maxiter": 10}
            )
            optimizer.optimize()

    def test_mixer_block_properties_sequence(self):
        """
        The custom mixers should have sequences that are correct.
        The sequence values are based on the position of the gate in the block
        relative to other gates of the same qubit count.
        """

        for (
            each_gatemap_list,
            each_gatemap_coeffs,
            correct_seq,
        ) in self.TESTING_GATEMAPS:
            gatemap_list_sequence = []
            one_qubit_count = 0
            two_qubit_count = 0

            for each_gatemap in each_gatemap_list:
                if each_gatemap.gate_label.n_qubits == 1:
                    gatemap_list_sequence.append(one_qubit_count)
                    one_qubit_count += 1

                elif each_gatemap.gate_label.n_qubits == 2:
                    gatemap_list_sequence.append(two_qubit_count)
                    two_qubit_count += 1

            # Test Equality between hand-written and programmatic assignment
            self.assertEqual(gatemap_list_sequence, correct_seq)

            qaoa_descriptor = QAOADescriptor(
                self.PROBLEM_QUBO.hamiltonian,
                each_gatemap_list,
                p=1,
                mixer_coeffs=each_gatemap_coeffs,
            )

            descriptor_mixer_seq = [
                each_mixer_gatemap.gate_label.sequence
                for each_mixer_gatemap in qaoa_descriptor.mixer_block
            ]

            # Test Equality between OQ and hand-written sequence
            self.assertEqual(descriptor_mixer_seq, correct_seq)

    def test_set_block_sequence(self):
        """
        Check that the set block sequence method is correct.
        """

        for (
            each_gatemap_list,
            each_gatemap_coeffs,
            correct_seq,
        ) in self.TESTING_GATEMAPS:
            output_gatemap_list = QAOADescriptor.set_block_sequence(each_gatemap_list)
            output_seq = [
                each_gatemap.gate_label.sequence for each_gatemap in output_gatemap_list
            ]

            self.assertEqual(output_seq, correct_seq)

    def test_set_block_sequence_error_raises(self):
        """
        Check that the set block sequence method raises the right error when the
        wrong type is passed. A TypeError should be raised if the input_gatemap_list
        if not a List of RotationGateMap Objects.
        """

        incorrect_input_iterable = [""], [1], [SWAPGateMap(0, 1)], [RXGateMap]

        for each_iterable in incorrect_input_iterable:
            with self.assertRaises(TypeError):
                QAOADescriptor.set_block_sequence(each_iterable)

    def test_block_setter(self):
        """
        Check that block_setter method correctly maps the sequence and the type
        of the RotationGateMap Objects returned.
        """

        input_enum_type = GateMapType.MIXER

        for (
            each_gatemap_list,
            each_gatemap_coeffs,
            correct_seq,
        ) in self.TESTING_GATEMAPS:
            output_gatemap_list = QAOADescriptor.block_setter(
                each_gatemap_list, input_enum_type
            )

            output_type = [
                each_gatemap.gate_label.type for each_gatemap in output_gatemap_list
            ]
            output_seq = [
                each_gatemap.gate_label.sequence for each_gatemap in output_gatemap_list
            ]

            # sequence and type should be labelled correctly.
            self.assertEqual(output_seq, correct_seq)
            self.assertEqual(
                output_type, [input_enum_type for _ in range(len(correct_seq))]
            )

    def test_block_setter_error_raises(self):
        """
        The block_setter method should raise a ValueError if the input_object is
        not of the type Hamiltonian or List. It should also raise a TypeError if
        the List contains elements that are not of the type RotationGateMap.
        """

        incorrect_input_object = [(), "", 1]

        for each_input in incorrect_input_object:
            with self.assertRaises(ValueError):
                QAOADescriptor.block_setter(
                    input_object=each_input, block_type=GateMapType.MIXER
                )

        incorrect_input_iterable = [
            [SWAPGateMap(0, 1)],
            [RXGateMap],
            [RXGateMap(0), SWAPGateMap(0, 1)],
        ]

        for each_input_iterable in incorrect_input_iterable:
            with self.assertRaises(TypeError):
                QAOADescriptor.block_setter(
                    input_object=each_input_iterable, block_type=GateMapType.MIXER
                )

    def test_block_setter_equivalence_simple(self):
        """
        A Hamiltonian Object and a list of RotationGateMap should have both their
        sequence and type assigned the same if they represent the same gate sequence.
        """

        # 1-Qubit
        test_hamiltonian = X_mixer_hamiltonian(3)
        test_gatemap_list = [RXGateMap(0), RXGateMap(1), RXGateMap(2)]

        return_gatemap_list_h = QAOADescriptor.block_setter(
            input_object=test_hamiltonian, block_type=GateMapType.MIXER
        )
        return_gatemap_list_gl = QAOADescriptor.block_setter(
            input_object=test_gatemap_list, block_type=GateMapType.MIXER
        )

        # Both gatemap list should be equivalent
        self.assertEqual(
            [
                each_gatemap.gate_label.sequence
                for each_gatemap in return_gatemap_list_h
            ],
            [
                each_gatemap.gate_label.sequence
                for each_gatemap in return_gatemap_list_gl
            ],
        )
        self.assertEqual(
            [each_gatemap.gate_label.type for each_gatemap in return_gatemap_list_h],
            [each_gatemap.gate_label.type for each_gatemap in return_gatemap_list_gl],
        )

        # 2-Qubit
        test_hamiltonian = XY_mixer_hamiltonian(3)
        test_gatemap_list = [
            RXXGateMap(0, 1),
            RXXGateMap(0, 2),
            RXXGateMap(1, 2),
            RYYGateMap(0, 1),
            RYYGateMap(0, 2),
            RYYGateMap(1, 2),
        ]

        return_gatemap_list_h = QAOADescriptor.block_setter(
            input_object=test_hamiltonian, block_type=GateMapType.MIXER
        )
        return_gatemap_list_gl = QAOADescriptor.block_setter(
            input_object=test_gatemap_list, block_type=GateMapType.MIXER
        )

        # Both gatemap list should be equivalent
        self.assertEqual(
            [
                each_gatemap.gate_label.sequence
                for each_gatemap in return_gatemap_list_h
            ],
            [
                each_gatemap.gate_label.sequence
                for each_gatemap in return_gatemap_list_gl
            ],
        )
        self.assertEqual(
            [each_gatemap.gate_label.type for each_gatemap in return_gatemap_list_h],
            [each_gatemap.gate_label.type for each_gatemap in return_gatemap_list_gl],
        )


if __name__ == "__main__":
    unittest.main()



import networkx as nx
import numpy as np
import scipy

from ..utilities import check_kwargs
from .problem import Problem
from .vehiclerouting import VRP
from .qubo import QUBO


class TSP(Problem):
    """
    The Traveling Salesman Problem (TSP) requires to find, given a list of cities and
    the distances between each pair of cities (or the cities coordinates),
    the shortest possible path that visits each city exactly once and returns to the origin city.
    Additionally, one can also specify how cities are connected together.
    Our implementation accepts three different kind of inputs:

    #. A list of the cities' coordinates and, optionally, a (directed) graph specifiying the connectivity between cities
    #. A distance matrix encoding distances between each pair of cities and, optionally, a (directed) graph specifiying the connectivity between cities
    #. A weighted (directed) graph specifiying the connectivity and the distance between cities

    Initializes a TSP object via three different methods:

    #. Give a list of coordinates for the cities and optionally the connectivity between them via a (directed) graph.
    #. Give a distance matrix and optionally the connectivity between cities via a (directed) graph.
    #. Directly give a (directed) weighted graph, where edge weights are interpreted as distances between cities

    Whenever no graph connectivity is specified, it is assumed that all cities are connected.

    Parameters
    ----------
    city_coordinates : Optional[List[Tuple[float, float]]]
        List containing the coordinates of each city.
    distance_matrix : Optional[List[List[float]]]
        Distance between cities given as list of list representing a matrix
    G: Optional[nx.Graph]
        Graph encoding the connectivity between cities (can be directed)
    A: Optional[float]
        Quadratic penalty coefficient to enforce that a path is a Hamiltonian cycle.
    B: Optional[float]
        Penalty coefficient which accounts for the path cost.

    Returns
    -------
    None
    """

    __name__ = "tsp"

    def __init__(
        self,
        city_coordinates=None,
        distance_matrix=None,
        G=None,
        A=None,
        B=1,
    ):
        # Initialization when a weighted graph is given
        if G is not None and nx.is_weighted(G):
            TSP.validate_graph(G)
            n_cities = len(G)
        else:
            # Initialization when cities coordinates are given
            if city_coordinates is not None:
                TSP.validate_coordinates(city_coordinates)
                n_cities = len(city_coordinates)
                distance_matrix = scipy.spatial.distance_matrix(
                    city_coordinates, city_coordinates
                )
            # Initialization when a distance matrix is given
            elif distance_matrix is not None:
                TSP.validate_distance_matrix(distance_matrix)
                n_cities = len(distance_matrix)
                distance_matrix = np.array(distance_matrix)
            # Raise error if no input is given
            else:
                raise ValueError(
                    "Input missing: city coordinates, distance matrix or (weighted graph) required"
                )

            # Take into account graph connectivity if unweighted graph is given
            G = G if G else nx.complete_graph(n_cities)
            if n_cities != len(G):
                raise ValueError(
                    "Number of cities does not match the number of nodes in graph"
                )

            # Set edge weights to be the distances between corresponding cities
            for u, v in G.edges():
                G[u][v]["weight"] = distance_matrix[u, v]

        # Set number of cities
        self.n_cities = n_cities

        # Set the graph, making sure it is directed (useful when looping over edges during QUBO creation)
        self._G = nx.DiGraph(G)

        # Set penalty coefficients if given, otherwise give default value
        self.A = A if A else 2 * distance_matrix.max()
        self.B = B

    @property
    def graph(self):
        return self._G

    @staticmethod
    def validate_coordinates(city_coordinates):
        """
        Makes the necessary check given some city coordinates.

        Parameters
        ----------
        input_coordinates : List[Tuple[float, float]]
            List containing the coordinates of each city.

        Returns
        -------
            None
        """
        if not isinstance(city_coordinates, list):
            raise TypeError("The coordinates should be a list")

        for each_entry in city_coordinates:
            if not isinstance(each_entry, tuple):
                raise TypeError("The coordinates should be contained in a tuple")

            for each_value in each_entry:
                if not isinstance(each_value, float) and not isinstance(
                    each_value, int
                ):
                    raise TypeError("The coordinates must be of type float or int")

    @staticmethod
    def validate_distance_matrix(distance_matrix):
        """
        Makes the necessary check given some distance matrix.

        Parameters
        ----------
        distance_matrix : List[List[float]]
            Distance between cities given as list of list representing a matrix

        Returns
        -------
            None
        """
        if not isinstance(distance_matrix, list):
            raise TypeError("The distance matrix should be a list")

        for each_entry in distance_matrix:
            if not isinstance(each_entry, list):
                raise TypeError("Each row in the distance matrix should be a list")

            for each_value in each_entry:
                if not isinstance(each_value, float) and not isinstance(
                    each_value, int
                ):
                    raise TypeError(
                        "The distance matrix entries must be of type float or int"
                    )

                if each_value < 0:
                    raise ValueError("Distances should be positive")

    @staticmethod
    def validate_graph(G):
        """
        Makes the necessary check given some (weighted) graph.

        Parameters
        ----------
        G: nx.Graph
            Graph encoding the connectivity between cities (can be directed)

        Returns
        -------
            None
        """
        # Set edge weights to be the distances between corresponding cities
        for u, v, weight in G.edges(data="weight"):
            print(weight)
            if not isinstance(weight, float) and not isinstance(weight, int):
                raise TypeError("The edge weights must be of type float or int")

            if weight < 0:
                raise ValueError("Edge weights should be positive")

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Traveling Salesman problem with
        fully connected cities.

        Parameters
        ----------
        n_cities: int
            The number of cities in the TSP instance. This is a required keyword argument.

        Returns
        -------
            A random instance of the Traveling Salesman problem.
        """
        n_cities = check_kwargs(["n_cities"], [None], **kwargs)[0]

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        # Generate random coordinates in a box of size sqrt(n_cities) x sqrt(n_cities)
        box_size = np.sqrt(n_cities)
        city_coordinates = list(map(tuple, box_size * rng.random(size=(n_cities, 2))))
        return TSP(city_coordinates=city_coordinates)

    def terms_and_weights(self):
        """
        Returns the terms and weights used in the QUBO formulation of this TSP instance.
        The QUBO formulation used is the one presented in Section 7.2 in
        https://arxiv.org/pdf/1302.5843.pdf, and sets the first city to be visited to be
        the first city in order to reduce the number of variables.

        Returns
        -------
        Tuple[List[List[int]], List[float]]
            Tuple containing a list with the terms and a list with the corresponding weights.
        """

        # Constants (flags) useful for the helper function below
        ZERO_VALUED_VARIABLE = -2
        ONE_VALUED_VARIABLE = -1

        def get_variable_index(v, j):
            """
            Returns the actual configuration index given the two indices v (city) and j (step),
            to mirror the formulation given in https://arxiv.org/pdf/1302.5843.pdf. Whenever the
            city or step probed is the first one, it can also return a flag saying whether the
            variable is 0 (flag=-2) or 1 (flag=-1), since the first city is fixed to reduce the
            number of variables).
            """
            if j > self.n_cities + 1 or v > self.n_cities:
                raise ValueError("Index out of bounds")

            # Whenever the step is the first one (or n+1 equivalently)
            if j == 1 or j == self.n_cities + 1:
                # If the city is the first one, we have x_{1, 1} = 1
                if v == 1:
                    variable_index = ONE_VALUED_VARIABLE
                # Else we have x_{v, 1} = 0
                else:
                    variable_index = ZERO_VALUED_VARIABLE

            # When step j>1 is given
            else:
                # If first node, then x_{1, j} = 0
                if v == 1:
                    variable_index = ZERO_VALUED_VARIABLE
                # Else return the index corresponding to variable x_{v, j}
                else:
                    variable_index = (j - 2) * (self.n_cities - 1) + (v - 2)

            return variable_index

        # Init the various terms
        constant_term = 0
        single_terms = []
        interaction_terms = []

        # Constraints ensuring that a city only appears once in the cycle, and that there is only one city per step
        # (note that it was simplified to account that the first city is always city 1)
        constant_term += 2 * self.A * (self.n_cities - 1)

        for v in range(2, self.n_cities + 1):
            for j in range(2, self.n_cities + 1):
                single_terms.append(([get_variable_index(v, j)], -4 * self.A))

        for k in range(2, self.n_cities + 1):
            for l in range(2, self.n_cities + 1):
                for v in range(2, self.n_cities + 1):
                    interaction_terms.append(
                        ([get_variable_index(v, k), get_variable_index(v, l)], self.A)
                    )

        for j in range(2, self.n_cities + 1):
            for u in range(2, self.n_cities + 1):
                for v in range(2, self.n_cities + 1):
                    interaction_terms.append(
                        ([get_variable_index(u, j), get_variable_index(v, j)], self.A)
                    )

        # Constraint which penalizes going through edges which are not part of the graph
        for u, v in nx.complement(self.graph).edges():
            for j in range(1, self.n_cities + 1):
                interaction_terms.append(
                    (
                        [
                            get_variable_index(u + 1, j),
                            get_variable_index(v + 1, j + 1),
                        ],
                        self.A,
                    )
                )

        # Terms to account for the path cost
        for u, v in self.graph.edges():
            for j in range(1, self.n_cities + 1):
                interaction_terms.append(
                    (
                        [
                            get_variable_index(u + 1, j),
                            get_variable_index(v + 1, j + 1),
                        ],
                        self.B * self.graph[u][v]["weight"],
                    )
                )

        # Filtering linear and quadratic terms such that variables which are fixed (and have been flagged)
        # can be processed accordingly
        filtered_interaction_terms = []
        for interaction, weight in single_terms + interaction_terms:
            # If the term is non-zero (so no flag=-2 is present), we should consider it
            if ZERO_VALUED_VARIABLE not in interaction:
                # If the same variable appears in a quadratic term, it becomes a linear term
                if len(interaction) == 2 and interaction[0] == interaction[1]:
                    interaction.pop()

                # Update interaction to reduce the degree of a term if some variables are set to 1
                # (that is remove all flag=-1)
                interaction = list(
                    filter(lambda a: a != ONE_VALUED_VARIABLE, interaction)
                )

                # Add the updated term
                filtered_interaction_terms.append((interaction, weight))

        # Unzip to retrieve terms and weights in separate sequences
        return tuple(zip(*(filtered_interaction_terms + [([], constant_term)])))

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        n = (self.n_cities - 1) ** 2
        terms, weights = self.terms_and_weights()

        # Convert to Ising equivalent since variables are in {0, 1} rather than {-1, 1}
        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(n, terms, weights)
        return QUBO(n, ising_terms, ising_weights, self.problem_instance)


class TSP_LP(VRP):
    """
    Creates an instance of the traveling salesman problem (TSP) based on the
    linear programming (LP) formulation. Note that the LP formulation of the TSP
    can be seen as a particular case of the Vehicle routing problem (VRP) with one vehicle.
    https://en.wikipedia.org/wiki/Travelling_salesman_problem


    Parameters
    ----------
    G: networkx.Graph
        Networkx graph of the problem
    pos: list[list]
        position x, y of each node
    subtours: list[list]
        if -1 (Default value): All the possible subtours are added to the constraints. Avoid it for large instances.
        if there are subtours that want be avoided in the solution, e.g, a 8 cities
        TSP with an optimal solution showing subtour between nodes 4, 5, and 8. It can be
        avoided introducing the constraint subtours=[[4,5,8]]. Additional information
        about subtours refer to https://de.wikipedia.org/wiki/Datei:TSP_short_cycles.png
    penalty: float
        Constraints penalty factor. If the method is 'unbalanced' three values are needed,
        one for the equality constraints and two for the inequality constraints.
    method: str
        Two available methods for the inequality constraints ["slack", "unbalanced"]
        For 'unblanced' see https://arxiv.org/abs/2211.13914
    Returns
    -------
        An instance of the TSP problem for the linear programming formulation.
    """

    __name__ = "tsp_lp"

    def __init__(self, G, pos, subtours, penalty, method):
        super().__init__(
            G, pos=pos, n_vehicles=1, subtours=subtours, method=method, penalty=penalty
        )

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the traveling salesman problem, whose graph is
        random.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:
            n_nodes: int
                The number of nodes (vertices) in the graph.
            method: str
                method for the inequality constraints ['slack', 'unbalanced'].
                For the unbalaced method refer to https://arxiv.org/abs/2211.13914
                For the slack method: https://en.wikipedia.org/wiki/Slack_variable
            subtours: list[list]
                Manually add the subtours to be avoided
            seed: int
                Seed for the random problems.

        Returns
        -------
            A random instance of the vehicle routing problem.
        """
        n_nodes = kwargs.get("n_nodes", 6)
        seed = kwargs.get("seed", None)
        method = kwargs.get("method", "slack")
        if method == "slack":
            penalty = kwargs.get("penalty", 4)
        elif method == "unbalanced":
            penalty = kwargs.get("penalty", [4, 1, 1])
        else:
            raise ValueError(f"The method '{method}' is not valid.")
        subtours = kwargs.get("subtours", -1)
        rng = np.random.default_rng(seed)
        G = nx.Graph()
        G.add_nodes_from(range(n_nodes))
        pos = [[0, 0]]
        pos += [list(2 * rng.random(2) - 1) for _ in range(n_nodes - 1)]
        for i in range(n_nodes - 1):
            for j in range(i + 1, n_nodes):
                r = np.sqrt((pos[i][0] - pos[j][0]) ** 2 + (pos[i][1] - pos[j][1]) ** 2)
                G.add_weighted_edges_from([(i, j, r)])

        return TSP_LP(G, pos=pos, subtours=subtours, method=method, penalty=penalty)

    def get_distance(self, sol):
        """
        Get the TSP solution distance

        Parameters
        ----------
        sol : str, dict
            Solution of the TSP problem.

        Returns
        -------
        total_distance : float
            Total distance of the current solution.

        """
        cities = self.G.number_of_nodes()
        if isinstance(sol, str):
            solution = {}
            for n, var in enumerate(self.docplex_model.iter_binary_vars()):
                solution[var.name] = int(sol[n])
            sol = solution
        total_distance = 0
        for i in range(cities):
            for j in range(i + 1, cities):
                if sol[f"x_{i}_{j}"]:
                    total_distance += self.G.edges[i, j]["weight"]
        return total_distance



from typing import List
import networkx as nx

from .basedevice import DeviceBase
from .plugin_finder import PLUGIN_DICT


def obtain_support_simulators() -> list:
    """This method returns a list containing the names of simulators supported
    by OQ. This list varies depending on the plugins available.
    """

    final_simulator_list = []
    for each_plugin_name, each_plugin_value in PLUGIN_DICT.items():
        if hasattr(each_plugin_value, "device_name_to_obj"):
            simulator_names = [
                each_key for each_key, _ in each_plugin_value.device_name_to_obj.items()
            ]
            final_simulator_list.extend(simulator_names)

    return final_simulator_list


SUPPORTED_LOCAL_SIMULATORS = [
    "vectorized",
    "analytical_simulator",
] + obtain_support_simulators()


class DeviceLocal(DeviceBase):
    """
    This class is a placeholder for all locally accessible devices.
    """

    def __init__(self, device_name: str):
        self.device_name = device_name
        self.device_location = "local"

    def check_connection(self) -> bool:
        if self.device_name in SUPPORTED_LOCAL_SIMULATORS:
            return True
        else:
            return False

    def connectivity(self, n_qubits: int) -> List[List[int]]:
        """
        The number of qubits for simulators depend on the problem
        """
        G = nx.complete_graph(n_qubits)
        return list(G.edges())



from typing import Union, List, Tuple
from abc import ABC, abstractmethod
from copy import deepcopy

import numpy as np
import networkx as nx


class UpdateMethod(ABC):
    @classmethod
    @abstractmethod
    def update(self):
        pass


class LoggerVariable(object):
    def __init__(
        self,
        attribute_name: str,
        history_update_method: UpdateMethod,
        best_update_method: UpdateMethod,
    ):
        self.name = attribute_name
        self.best = []
        self.history = []

        self.history_update_method = history_update_method
        self.best_update_method = best_update_method

    def update(self, new_value):
        self.update_history(new_value)
        self.update_best(new_value)

    def update_history(self, new_value):
        return self.history_update_method.update(self, "history", new_value)

    def update_best(self, new_value):
        return self.best_update_method.update(self, "best", new_value)


class AppendValue(UpdateMethod):
    def update(
        self, attribute_name: str, new_value: Union[list, int, float, str]
    ) -> None:
        get_var = getattr(self, attribute_name)
        get_var.append(new_value)
        setattr(self, attribute_name, get_var)


class ReplaceValue(UpdateMethod):
    def update(
        self, attribute_name: str, new_value: Union[list, int, float, str]
    ) -> None:
        setattr(self, attribute_name, [new_value])


class EmptyValue(UpdateMethod):
    def update(
        self, attribute_name: str, new_value: Union[list, int, float, str]
    ) -> None:
        setattr(self, attribute_name, [])


class IfLowerDo(UpdateMethod):
    def __init__(self, update_method: UpdateMethod):
        self._update_method = update_method

    def update(
        self,
        logger_variable: LoggerVariable,
        attribute_name: str,
        new_value: Union[int, float],
    ) -> None:
        try:
            old_value = getattr(logger_variable, attribute_name)[-1]
            if new_value < old_value:
                self._update_method.update(logger_variable, attribute_name, new_value)
        except IndexError:
            AppendValue.update(logger_variable, attribute_name, new_value)


class IfHigherDo(UpdateMethod):
    def __init__(self, update_method: UpdateMethod):
        self._update_method = update_method

    def update(
        self,
        logger_variable: LoggerVariable,
        attribute_name: str,
        new_value: Union[int, float],
    ) -> None:
        try:
            old_value = getattr(logger_variable, attribute_name)[-1]
            if new_value > old_value:
                self._update_method.update(logger_variable, attribute_name, new_value)
        except IndexError:
            AppendValue.update(logger_variable, attribute_name, new_value)


class LoggerVariableFactory(object):

    """
    Creates Logger Variable Objects
    """

    history_bool_mapping = {"True": AppendValue, "False": EmptyValue}
    best_string_mapping = {
        "HighestOnly": IfHigherDo(ReplaceValue),
        "HighestSoFar": IfHigherDo(AppendValue),
        "LowestOnly": IfLowerDo(ReplaceValue),
        "LowestSoFar": IfLowerDo(AppendValue),
        "Replace": ReplaceValue,
        "Append": AppendValue,
    }

    @classmethod
    def create_logger_variable(
        self, attribute_name: str, history_update_bool: bool, best_update_string: str
    ) -> LoggerVariable:
        history_update_method = self.history_bool_mapping[str(history_update_bool)]
        best_update_method = self.best_string_mapping[best_update_string]

        return LoggerVariable(attribute_name, history_update_method, best_update_method)


class Logger(object):
    def __init__(self, initialisation_variables: dict, logger_update_structure: dict):
        for attribute_name, attribute_properties in initialisation_variables.items():
            self._create_variable(
                attribute_name=attribute_name,
                history_update_bool=attribute_properties["history_update_bool"],
                best_update_string=attribute_properties["best_update_string"],
            )
        self._create_logger_update_structure(
            logger_update_structure["root_nodes"],
            logger_update_structure["best_update_structure"],
        )

    def _create_variable(
        self, attribute_name: str, history_update_bool: bool, best_update_string: str
    ):
        """
        If a variable was created was this method, ensure that the logger update
        strucuture is updated to include the new variable. Else the best value
        of the new variable will not be updated.
        """

        setattr(
            self,
            attribute_name,
            LoggerVariableFactory.create_logger_variable(
                attribute_name, history_update_bool, best_update_string
            ),
        )

    def _create_logger_update_structure(
        self, root_nodes: List[str], best_update_relations: Tuple[List[str]]
    ):
        """
        This method creates an internal representation based on the way in which
        the Logger Variables should be updated using a NetworkX Directed Graph.

        Parameter
        ---------
        root_nodes: `list`
            Starting points of the update graphs. The name of the Logger Variable
            that needs to be updated first should be named in this list.
        best_update_relations: `list`
            If the updating of B depends on A, (A changes, Update B). This
            relationship should be encapsuled/represented with the following
            list: ['A', 'B'].
        """

        self.best_update_structure = nx.DiGraph()

        self.root_nodes = root_nodes
        self.update_edges = best_update_relations

    @property
    def update_edges(self):
        return self._update_edges

    @update_edges.setter
    def update_edges(self, update_edges: Tuple[List[str]]):
        self.best_update_structure.add_edges_from(update_edges)
        self._update_edges = update_edges

        return self._update_edges

    @property
    def root_nodes(self):
        return self._root_nodes

    @root_nodes.setter
    def root_nodes(self, root_nodes: List[str]):
        self.best_update_structure.add_nodes_from(root_nodes)
        self._root_nodes = root_nodes

        return self._root_nodes

    def log_variables(self, input_dict: dict):
        """
        Updates all the histories of the logger variables.
        Best values of the logger variables are only updated according to the rules
        specified by its update structure. If the attribute is not in the update
        structure it isnt updated.

        input_dict:
            Should contain a dictionary with the key as the name of the
            attribute to be updated and its respective value. Note that the
            attribute should have been created beforehand.
        """

        for each_key in input_dict.keys():
            self._log_history(each_key, input_dict[each_key])

        # Updates based on best_update_structure, the networkx graph.
        change_dict = dict()

        node_list = deepcopy(self.root_nodes)

        while len(node_list) != 0:
            mid_list = []
            for each_node in node_list:
                change_bool = False

                # Checks if a nodes predecessor has been updated. If not all
                # of them were updated, the node is skipped.
                node_predecessor = self.best_update_structure.pred[each_node]
                updated_pred_count = np.sum(
                    [
                        change_dict[each_precessor_node]
                        for each_precessor_node in node_predecessor
                    ]
                )
                if len(node_predecessor) != updated_pred_count:
                    continue

                # Update node if attribute name is in input dictionary and
                # its predecessors hsa been updated.
                if each_node in input_dict.keys():
                    logged_var = getattr(self, each_node)
                    old_best = logged_var.best.copy()
                    self._log_best(each_node, input_dict[each_node])
                    new_best = logged_var.best
                    if type(new_best[0]) == np.ndarray:
                        if not np.array_equal(new_best, old_best):
                            change_bool = True
                    elif new_best != old_best:
                        change_bool = True

                # Keeps track of the nodes whose best values have changed
                change_dict[each_node] = change_bool

                # Retrieve the next set of nodes to update if the current node
                # was changed.
                if change_bool == True:
                    mid_list.extend(self.best_update_structure.adj[each_node].keys())

            node_list = list(set(mid_list))

    def _log_history(self, attribute_name: str, attribute_value):
        attr_to_update = getattr(self, attribute_name)
        attr_to_update.update_history(attribute_value)

    def _log_best(self, attribute_name: str, attribute_value):
        attr_to_update = getattr(self, attribute_name)
        attr_to_update.update_best(attribute_value)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# # NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""
This package contains unified functions for framework-agnostic tensor and array
manipulation. Given the input tensor-like object, the call is dispatched
to the corresponding array manipulation framework, allowing for end-to-end
differentiation to be preserved.

.. warning::

    These functions are experimental, and only a subset of common functionality is supported.
    Furthermore, the names and behaviour of these functions may differ from similar
    functions in common frameworks; please refer to the function docstrings for more details.

The following frameworks are currently supported:

* NumPy
"""  # changed from the original PennyLane code

import autoray as ar

from .multi_dispatch import (
    _multi_dispatch,
    multi_dispatch,
    array,
    block_diag,
    concatenate,
    diag,
    dot,
    einsum,
    eye,
    frobenius_inner_product,
    get_trainable_indices,
    ones_like,
    scatter,
    scatter_element_add,
    stack,
    tensordot,
    unwrap,
    where,
    add,
    iscomplex,
    expm,
)

from .quantum import cov_matrix, marginal_prob
from .quantum import (
    reduced_dm,
    vn_entropy,
    mutual_info,
    sqrt_matrix,
    fidelity,
    relative_entropy,
)

from .utils import (
    allclose,
    allequal,
    cast,
    cast_like,
    in_backprop,
    is_abstract,
    convert_like,
    get_interface,
    requires_grad,
)

from .is_independent import is_independent

from .matrix_manipulation import expand_matrix, reduce_matrices

sum = ar.numpy.sum
toarray = ar.numpy.to_numpy
T = ar.numpy.transpose


# small constant for numerical stability that the user can modify
eps = 1e-14


def __getattr__(name):
    return getattr(ar.numpy, name)


__all__ = [
    "_multi_dispatch",
    "multi_dispatch",
    "allclose",
    "allequal",
    "array",
    "block_diag",
    "cast",
    "cast_like",
    "concatenate",
    "convert_like",
    "cov_matrix",
    "diag",
    "dot",
    "einsum",
    "eye",
    "fidelity",
    "frobenius_inner_product",
    "get_interface",
    "get_trainable_indices",
    "in_backprop",
    "is_abstract",
    "is_independent",
    "marginal_prob",
    "mutual_info",
    "ones_like",
    "reduced_dm",
    "relative_entropy",
    "requires_grad",
    "sqrt_matrix",
    "scatter_element_add",
    "stack",
    "tensordot",
    "unwrap",
    "vn_entropy",
    "where",
    "add",
    "iscomplex",
    "expand_matrix",
]
